import atexit
import os
import shlex
import subprocess
import tempfile

from cucu import config, logger, step


def run_command(
    command,
    stdout_var=None,
    stderr_var=None,
    exit_code_var=None,
    check_exit_code=None,
):
    args = shlex.split(command)
    process = subprocess.run(args, capture_output=True)

    if exit_code_var:
        config.CONFIG[exit_code_var] = str(process.returncode)

    stdout = process.stdout.decode("utf8")
    stderr = process.stderr.decode("utf8")

    if stdout_var:
        config.CONFIG[stdout_var] = config.CONFIG.escape(stdout)

    if stderr_var:
        config.CONFIG[stderr_var] = config.CONFIG.escape(stderr)

    return_code = process.returncode
    if check_exit_code is not None and int(check_exit_code) != return_code:
        logger.error(f"STDOUT:\n{stdout}\n")
        logger.error(f"STDERR:\n{stderr}\n")
        raise RuntimeError(
            f"expected exit code {check_exit_code}, got {return_code}, see above for details"
        )

    else:
        logger.debug(f"STDOUT:\n{stdout}\n")
        logger.debug(f"STDERR:\n{stderr}\n")


def run_script(
    script,
    stdout_var=None,
    stderr_var=None,
    exit_code_var=None,
    check_exit_code=None,
):
    script_fd, script_filename = tempfile.mkstemp()
    os.close(script_fd)
    atexit.register(os.remove, script_filename)

    with open(script_filename, "wb") as script_file:
        script_file.write(script.encode())

    os.chmod(script_filename, 0o755)
    process = subprocess.run(
        [script_file.name],
        capture_output=True,
        shell=True,
    )  # nosec

    if exit_code_var:
        config.CONFIG[exit_code_var] = str(process.returncode)

    stdout = process.stdout.decode("utf8")
    stderr = process.stderr.decode("utf8")

    if stdout_var:
        config.CONFIG[stdout_var] = config.CONFIG.escape(stdout)

    if stderr_var:
        config.CONFIG[stderr_var] = config.CONFIG.escape(stderr)

    return_code = process.returncode
    if check_exit_code is not None and int(check_exit_code) != return_code:
        logger.error(f"STDOUT:\n{stdout}\n")
        logger.error(f"STDERR:\n{stderr}\n")
        raise RuntimeError(
            f"expected exit code {check_exit_code}, got {return_code}, see above for details"
        )
    else:
        logger.debug(f"STDOUT:\n{stdout}\n")
        logger.debug(f"STDERR:\n{stderr}\n")


@step(
    'I run the command "{command}" and save stdout to "{stdout_var}", stderr to "{stderr_var}", exit code to "{exit_code_var}"'
)
def runs_command_and_save_everything(
    context, command, stdout_var, stderr_var, exit_code_var
):
    run_command(
        command,
        stdout_var=stdout_var,
        stderr_var=stderr_var,
        exit_code_var=exit_code_var,
    )


@step(
    'I run the command "{command}" and save stdout to "{stdout_var}", stderr to "{stderr_var}" and expect exit code "{exit_code}"'
)
def runs_command_and_check_exit_code(
    context, command, stdout_var, stderr_var, exit_code
):
    run_command(
        command,
        stdout_var=stdout_var,
        stderr_var=stderr_var,
        check_exit_code=exit_code,
    )


@step(
    'I run the command "{command}" and save stdout to "{stdout_var}" and expect exit code "{exit_code}"'
)
def runs_command__and_save_stdout_and_check_exit_code(
    context, command, stdout_var, exit_code
):
    run_command(
        command,
        stdout_var=stdout_var,
        check_exit_code=exit_code,
    )


@step(
    'I run the command "{command}" and save stdout to "{stdout_var}", exit code to "{exit_code_var}"'
)
def run_command_and_save_stdout_and_exit_code(
    context, command, stdout_var, exit_code_var
):
    run_command(command, stdout_var=stdout_var, exit_code_var=exit_code_var)


@step('I run the command "{command}" and save exit code to "{exit_code_var}"')
def run_command_and_save_exit_code(context, command, exit_code_var):
    run_command(command, exit_code_var=exit_code_var)


@step('I run the command "{command}" and expect exit code "{exit_code}"')
def run_command_and_expect_exit_code(context, command, exit_code):
    run_command(command, check_exit_code=exit_code)


@step(
    'I run the following script and save stdout to "{stdout_var}", stderr to "{stderr_var}", exit code to "{exit_code_var}"'
)
def run_script_and_save_everything(
    context, stdout_var, stderr_var, exit_code_var
):
    run_script(
        context.text,
        stdout_var=stdout_var,
        stderr_var=stderr_var,
        exit_code_var=exit_code_var,
    )


@step(
    'I run the following script and save stdout to "{stdout_var}", stderr to "{stderr_var}" and expect exit code "{exit_code}"'
)
def run_script_and_check_exit_code(context, stdout_var, stderr_var, exit_code):
    run_script(
        context.text,
        stdout_var=stdout_var,
        stderr_var=stderr_var,
        check_exit_code=exit_code,
    )


@step('I run the following script and expect exit code "{exit_code}"')
def run_script_and_expect_exit_code(context, exit_code):
    run_script(
        context.text,
        check_exit_code=exit_code,
    )



import json
import logging
import os
import re
import socket
from pathlib import Path

import yaml


class Config(dict):
    # only match variables {...}
    __VARIABLE_REGEX = re.compile(r"\{(?<!\\{)([^{}]+)\}(?<!\\})")

    def __init__(self, **kwargs):
        self.update(**kwargs)
        self.resolving = False
        self.defined_variables = {}
        self.variable_lookups = {}
        self.snapshots = []

    def define(self, name, description, default=None):
        """
        used to define variables and set their default so that we can then
        provide the info using the `cucu vars` command for test writers to know
        where and how to use various built-in variables.

        parameters:
            name(string): the name of the variable
            description(string): a succinct description of variables purpose
            default(string): an optional default to set the variable to
        """
        self.defined_variables[name] = {
            "description": description,
            "default": default,
        }
        # set the default
        self.__setitem__(name, default)

    def escape(self, string):
        """
        utility method used to escape strings that would be otherwise problematic
        if the values were passed around as is in cucu steps.
        """
        if isinstance(string, str):
            string = string.replace("{", "\\{")
            string = string.replace("}", "\\}")

        return string

    def __getitem__(self, key):
        try:
            # environment always takes precedence
            value = os.environ.get(key)

            if value is None:
                value = dict.__getitem__(self, key)

            return value
        except KeyError:
            return None

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)

    def get(self, key, default=None):
        try:
            for regex, lookup in self.variable_lookups.items():
                if regex.match(key):
                    return lookup(key)

            return self[key]
        except KeyError:
            return default

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def bool(self, key):
        return self[key] in [True, "True", "true", "yes", "enabled"]

    def true(self, key):
        return self.bool(key)

    def false(self, key):
        return not self.true(key)

    def load(self, filepath):
        """
        loads configuration values from a YAML file at the filepath provided
        """
        config = yaml.safe_load(open(filepath, "rb"))

        if config:
            for key in config.keys():
                if key == "CUCU_SECRETS":  # security: only add restrictions
                    vals = self.get(key, "").split(",")
                    vals = vals + config[key].split(",")
                    self[key] = ",".join(set(vals) ^ {""})
                else:
                    self[key] = config[key]

    def load_cucurc_files(self, filepath: Path):
        """
        load in order the ~/.cucurc.yml and then subsequent config files
        starting from the current working directory to the filepath provided
        """

        # load the ~/.cucurc.yml first
        home_cucurc_filepath = os.path.join(
            os.path.expanduser("~"), ".cucurc.yml"
        )
        if os.path.exists(home_cucurc_filepath):
            self.load(home_cucurc_filepath)

        filepath = os.path.abspath(filepath)
        if os.path.isfile(filepath):
            basename = os.path.dirname(filepath)
        else:
            basename = filepath

        # create the inverse list of the directories starting from cwd to the one
        # containing the feature file we want to run to load the cucrc.yml files in
        # the correct order
        dirnames = [os.getcwd()]
        while basename != os.getcwd():
            dirnames.append(basename)
            basename = os.path.dirname(basename)

        for dirname in dirnames:
            cucurc_filepath = os.path.join(dirname, "cucurc.yml")
            if os.path.exists(cucurc_filepath):
                self.load(cucurc_filepath)

    def expand(self, string):
        """
        find any variable references and return an dictionary of all of the
        variable names and values within the string provided.

        returns:
            a dictionary of all of the variable names found in the string
            provided with the exact value of the variable at runtime.
        """
        references = {}

        variables = re.findall("{([^{}]+)}", string)

        for variable in variables:
            value = self.resolve(f"{{{variable}}}")

            # if it didn't resolve to anything then
            if value:
                value = str(value).replace("\n", "\\n")
                value = value[:80] + "..." * (len(value) > 80)
            else:
                value = None

            references[variable] = value

        return references

    def hide_secrets(self, text: str | bytes):
        secret_keys = [x for x in self.get("CUCU_SECRETS", "").split(",") if x]
        secret_values = [self.get(x) for x in secret_keys if self.get(x)]
        secret_values = [x for x in secret_values if isinstance(x, str)]

        is_bytes = isinstance(text, bytes)
        if is_bytes:
            text = text.decode()

        result = None
        if text.startswith("{"):
            try:
                result = self._hide_secrets_json(secret_values, text)
            except Exception as e:
                print(
                    f"Couldn't parse json, falling back to text processing: {e}"
                )

        if result is None:
            result = self._hide_secrets_text(secret_values, text)

        if is_bytes:
            result = result.encode()

        return result

    def _hide_secrets_json(self, secret_values, text):
        data = json.loads(text)

        def hide_node(value, parent, key):
            if not isinstance(value, str):
                return value

            if (
                key == "name"
                and isinstance(parent, dict)
                and parent.get("keyword", "") in ["Feature", "Scenario"]
            ):
                return value

            return self._hide_secrets_text(secret_values, value)

        leaf_map(data, hide_node)
        return json.dumps(data, indent=2, sort_keys=True)

    def _hide_secrets_text(self, secret_values, text):
        lines = text.split("\n")

        for x in range(len(lines)):
            # here's where we can hide secrets
            for value in secret_values:
                lines[x] = lines[x].replace(value, "*" * len(value))

        return "\n".join(lines)

    def resolve(self, string):
        """
        resolve any variable references {...} in the string provided using
        the values currently stored in the config object. We also unescape any
        characters preceding by a backslash as to allow for test writers to
        escape special characters
        """
        if isinstance(string, str):
            previouses = [None]
            while previouses[-1] != string:
                # if any of the previous iterations of replacing variables looks
                # like the exact string we have now then break out of the loop
                # as there's an infinite recursion of variable replacing values
                if string in previouses:
                    raise RuntimeError("infinite replacement loop detected")

                previouses.append(string)

                for var_name in Config.__VARIABLE_REGEX.findall(string):
                    value = self.get(var_name)

                    if value is None:
                        value = ""
                        # print directly to the output stream, which was taken over in behave_tweaks
                        print(f'WARNING variable "{var_name}" is undefined')

                    string = string.replace("{" + var_name + "}", str(value))

            # we are only going to allow escaping of { and " characters for the
            # time being as they're part of the language:
            #
            #   * " are used around step arguments
            #   * {} are used by variable references
            #
            # regex below uses negative lookbehind to assert we're not replacing
            # any escapes that have the backslash itself escaped.
            #
            string = re.sub(r"(?<!\\)\\{", "{", string)
            string = re.sub(r"(?<!\\)\\}", "}", string)
            string = re.sub(r'(?<!\\)\\"', '"', string)

        return string

    def snapshot(self, name=None):
        if name is None:
            name = f"snapshot_{len(self.snapshots)}"

        snapshot_data = {"name": name, "config": self.copy()}
        self.snapshots.append(snapshot_data)

        logging.debug(
            f"CONFIG: snapshot taken '{name}' (stack depth: {len(self.snapshots)})"
        )

    def restore(self, with_pop=False):
        if not self.snapshots:
            return

        if with_pop:
            latest_snapshot = self.snapshots.pop()
            action = "popped and restored"
        else:
            latest_snapshot = self.snapshots[-1]
            action = "restored to"

        self.clear()
        self.update(**latest_snapshot["config"])

        logging.debug(
            f"CONFIG: {action} snapshot '{latest_snapshot['name']}' (stack depth: {len(self.snapshots)})"
        )

    def list_snapshots(self):
        return [snapshot["name"] for snapshot in self.snapshots]

    def register_custom_variable_handling(self, regex, lookup):
        """
        register a regex to match variable names on and allow the lookup
        function provided to do the handling of the resolution of the variable
        name.

        parameters:
            regex(string): regular expression to match on any config variable
                           name when doing lookups
            lookup(func): a function that accepts a variable name to return its
                          value at runtime.

                          def lookup(name):
                            return [value of the variable at runtime]
        """
        self.variable_lookups[re.compile(regex)] = lookup

    def to_yaml_without_secrets(self):
        omit_keys = [
            "CWD",
            "STDOUT",
            "STDERR",
            "SCENARIO_DOWNLOADS_DIR",
            "SCENARIO_LOGS_DIR",
            "SCENARIO_RESULTS_DIR",
        ]
        secret_keys = [x for x in self.get("CUCU_SECRETS", "").split(",") if x]
        keys = [
            k
            for k in self.keys()
            if not k.startswith("__")
            and k not in secret_keys
            and k not in omit_keys
        ]

        config = {k: self[k] for k in sorted(keys)}
        for k, v in config.items():
            if isinstance(v, str):
                v = self.hide_secrets(v)
                # fix leading '*' issue
                if v.startswith("*"):
                    v = f"'{v}'"
                config[k] = v

        # Resolve non-atomic classes to their string representation
        yaml.representer.SafeRepresenter.add_representer(
            None,
            lambda dumper, data: dumper.represent_scalar(
                "tag:yaml.org,2002:str", str(data)
            ),
        )
        return yaml.safe_dump(config)


# global config object
CONFIG = Config()


def _get_local_address():
    """
    internal method to get the current host address
    """
    google_dns_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    google_dns_socket.connect(("8.8.8.8", 80))
    return google_dns_socket.getsockname()[0]


CONFIG.define(
    "HOST_ADDRESS",
    "host address of the current machine cucu is running on",
    default=_get_local_address(),
)
CONFIG.define(
    "CWD",
    "the current working directory of the cucu process",
    default=os.getcwd(),
)
CONFIG.define(
    "CUCU_SECRETS",
    "a comma separated list of VARIABLE names that we want to hide "
    "their values in the various outputs of cucu by replacing with asterisks",
    default="",
)
CONFIG.define(
    "CUCU_SHORT_UI_RETRY_AFTER_S",
    "the amount of time to wait between retries in seconds for non-wait ui steps",
    default=0.25,
)
CONFIG.define(
    "CUCU_SHORT_UI_WAIT_TIMEOUT_S",
    "the total amount of wait time in seconds for non-wait ui steps",
    default=2.25,
)
CONFIG.define(
    "CUCU_STEP_WAIT_TIMEOUT_S",
    "the total amount of wait time in seconds for `wait for` steps",
    default=20.0,
)
CONFIG.define(
    "CUCU_STEP_RETRY_AFTER_S",
    "the amount of time to wait between retries in seconds for `wait for` steps",
    default=0.5,
)
CONFIG.define(
    "CUCU_KEEP_BROWSER_ALIVE",
    "when set to true we'll reuse the browser between scenario runs",
    default=False,
)
CONFIG.define(
    "CUCU_BROWSER_WINDOW_HEIGHT",
    "the browser window height when running browser tests",
    default=768,
)
CONFIG.define(
    "CUCU_BROWSER_WINDOW_WIDTH",
    "the browser window width when running browser tests",
    default=1366,
)
CONFIG.define(
    "CUCU_BROWSER_DOWNLOADS_DIR",
    "the browser download directory when running browser tests",
    default="/tmp/cucu-browser-downloads",
)
CONFIG.define(
    "CUCU_SOCKET_DEFAULT_TIMEOUT_S",
    "the default timeout (seconds) for socket connect/read in cucu",
    default=10,
)
CONFIG.define(
    "CUCU_SELENIUM_DEFAULT_TIMEOUT_S",
    "the default timeout (seconds) for selenium connect/read in selenium",
    default=10,
)
CONFIG.define(
    "CUCU_MONITOR_PNG",
    "when set to a filename `cucu` will update the image to match "
    "the exact image step at runtime.",
    default=".monitor.png",
)
CONFIG.define(
    "CUCU_LINT_RULES_PATH",
    "comma separated list of paths to load cucu lint rules from .yaml files",
    default="",
)
CONFIG.define(
    "CUCU_JUNIT_WITH_STACKTRACE",
    "when set to 'true' results in stacktraces showing in the JUnit XML failure output",
    default="false",
)
CONFIG.define(
    "CUCU_SKIP_HIGHLIGHT_BORDER",
    "when set to 'True' skips adding a border to highlight found element in screenshots",
    default=True,
)


# define re_map here instead of in utils.py to avoid circular import
def leaf_map(data, value_func, parent=None, key=None):
    """
    Utility to apply a map function recursively to a dict or list.

    Args:
        data: The dict or list or value to use
        value_func: Callable function that accepts data and parent
        parent: The parent object (or None)
    """
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = leaf_map(value, value_func, data, key)
        return data
    elif isinstance(data, list):
        for x, value in enumerate(data):
            data[x] = leaf_map(value, value_func, data, key)
        return data
    else:
        return value_func(data, parent, key)



#
# menuitem steps
# https://www.w3.org/TR/wai-aria-1.1/#menuitem
#
from cucu import fuzzy, helpers
from cucu.utils import take_saw_element_screenshot

from . import base_steps


def find_menuitem(ctx, name, index=0):
    """
    find a menuitem containing the menuitem provide.

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired menuitem on screen
      index(str):  the index of the menuitem if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser, name, ['*[role="menuitem"]'], index=index
    )

    take_saw_element_screenshot(ctx, "menuitem", name, index, element)

    return element


helpers.define_should_see_thing_with_name_steps("menu item", find_menuitem)
helpers.define_thing_with_name_in_state_steps(
    "menu item", "disabled", find_menuitem, base_steps.is_disabled
)
helpers.define_thing_with_name_in_state_steps(
    "menu item", "not disabled", find_menuitem, base_steps.is_not_disabled
)



from cucu import fuzzy, helpers, step
from cucu.browser.frames import try_in_frames_until_success
from cucu.steps import step_utils
from cucu.utils import take_saw_element_screenshot, text_in_current_frame


def find_text(ctx, name, index=0):
    """
    find any element containing the text provide.

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired radio text on screen
      index(str):  the index of the radio text if there are duplicates

    returns:
        the WebElement that matches the provided arguments or None if none found
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser,
        name,
        ["*"],
        index=index,
        direction=fuzzy.Direction.LEFT_TO_RIGHT,
    )

    take_saw_element_screenshot(ctx, "text", name, index, element)

    return element


# Also update the line number in the  scenario: `User gets the right stacktrace for steps using step helpers` when changing the code below.
helpers.define_should_see_thing_with_name_steps("text", find_text)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "text", find_text
)


@step(
    'I search for the regex "{regex}" on the current page and save the group "{name}" to the variable "{variable}"'
)
def search_for_regex_to_page_and_save(ctx, regex, name, variable):
    ctx.check_browser_initialized()

    def search_for_regex_in_frame():
        text = text_in_current_frame(ctx.browser)
        step_utils.search_and_save(
            regex=regex, value=text, name=name, variable=variable
        )

    try_in_frames_until_success(ctx.browser, search_for_regex_in_frame)


@step(
    'I match the regex "{regex}" on the current page and save the group "{name}" to the variable "{variable}"'
)
def match_for_regex_to_page_and_save(ctx, regex, name, variable):
    ctx.check_browser_initialized()

    def match_for_regex_in_frame():
        text = text_in_current_frame(ctx.browser)
        step_utils.match_and_save(
            regex=regex, value=text, name=name, variable=variable
        )

    try_in_frames_until_success(ctx.browser, match_for_regex_in_frame)


@step('I should see text matching the regex "{regex}" on the current page')
def search_for_regex_on_page(ctx, regex):
    ctx.check_browser_initialized()

    def search_for_regex_in_frame():
        text = text_in_current_frame(ctx.browser)
        step_utils.search(regex=regex, value=text)

    try_in_frames_until_success(ctx.browser, search_for_regex_in_frame)



import pkgutil
import re
from io import StringIO

from selenium.webdriver.common.by import By

from cucu import (
    config,
    format_gherkin_table,
    fuzzy,
    helpers,
    logger,
    retry,
    step,
)
from cucu.browser.frames import run_in_all_frames


def find_tables(ctx):
    """
    find all the tables currently present on the page

    parameters:
      ctx(object): behave context object used to share data between steps

    returns:
      an array of arrays containing the HTML tables currently displayed
    """
    ctx.check_browser_initialized()
    tables_lib = pkgutil.get_data("cucu", "steps/tables.js")
    tables_lib = tables_lib.decode("utf8")

    def search_for_tables():
        ctx.browser.execute(tables_lib)
        return ctx.browser.execute("return findAllTables();")

    return run_in_all_frames(ctx.browser, search_for_tables)


def behave_table_to_array(table):
    """
    given a behave.model.Table object convert it to an array of rows

    parameters:
        table(behave.model.Table): the behave table to convert into an array

    returns:
        array of rows representing the behave table provided.
    """
    result = [table.headings]

    for row in table.rows:
        values = []
        for value in row:
            values.append(value)
        result.append(values)

    return result


def check_table_equals_table(table, expected_table):
    """
    check that table is equal to expected_table
    """
    return table == expected_table


def check_table_matches_table(table, expected_table):
    """
    check if table matches the regex patterns in expected table
    """

    table_matched = bool(len(table) == len(expected_table))
    if table_matched:
        for expected_row, row in zip(expected_table, table):
            for expected_value, value in zip(expected_row, row):
                if not re.match(expected_value, value):
                    table_matched = False
                    break

    return table_matched


def check_table_contains_table(table, expected_table):
    """
    check that table contains the rows in expected_table
    """
    return all(row in table for row in expected_table)


def check_table_contains_matching_rows_in_table(table, expected_table):
    """
    check that table contains the matching rows in expected_table
    """
    table_matched = True

    for expected_row in expected_table:
        for row in table:
            found_row = True
            for expected_value, value in zip(expected_row, row):
                if not re.match(expected_value, value):
                    found_row = False
            if found_row:
                break

        if not found_row:
            table_matched = False
            break

    if table_matched:
        return True

    return False


def report_unable_to_find_table(expected_table, found_tables):
    stream = StringIO()
    stream.write("\n")
    for index, table in enumerate(found_tables):
        print_index = helpers.nth_to_ordinal(index) or '"1st" '
        stream.write(
            f"{print_index}table:\n{format_gherkin_table(table, [], '  ')}\n"
        )

    stream.seek(0)
    raise RuntimeError(
        f"unable to find desired table\nexpected:\n{format_gherkin_table(expected_table, [], '  ')}\n\nfound:{stream.read()}"
    )


def report_found_undesired_table(unexpected_tables, found_tables):
    stream = StringIO()
    stream.write("\n")
    for index, table in enumerate(found_tables):
        print_index = helpers.nth_to_ordinal(index) or '"1st" '
        stream.write(
            f"{print_index}table:\n{format_gherkin_table(table, [], '  ')}\n"
        )

    stream.seek(0)
    error_message = ""
    for table in unexpected_tables:
        error_message += (
            f"found undesired table\n\nundesired table:\n{format_gherkin_table(table, [], '  ')}\n\n"
            f"all tables found:{stream.read()}\n"
        )
    raise RuntimeError(error_message)


def find_table(ctx, assert_func, nth=None):
    """
    validate we can find the table passed in the ctx object and assert it
    matches anyone of the tables on the current web page. If `nth` is set to
    something then we only check against the nth table of the available tables.

    paramters:
        ctx(object): behave context object
        assert_func(function): function used to assert two tables "match"
        nth(int): when set to an int specifies the exact table within the list
                  of available tables to match against.

    raises:
        RuntimeError when the desired table was not found
    """
    expected = behave_table_to_array(ctx.table)
    found_tables = find_tables(ctx)

    if nth is not None:
        if assert_func(found_tables[nth], expected):
            return

    else:
        for table in found_tables:
            if assert_func(table, expected):
                return

    report_unable_to_find_table(expected, found_tables)


def do_not_find_table(ctx, assert_func, nth=None):
    """
    validate we can not find the table passed in the ctx object and assert it
    matches anyone of the tables on the current web page. If `nth` is set to
    something then we only check against the nth table of the available tables.

    paramters:
        ctx(object): behave context object
        assert_func(function): function used to assert two tables "match"
        nth(int): when set to an int specifies the exact table within the list
                  of available tables to match against.

    raises:
        RuntimeError when the desired table was not found
    """
    expected = behave_table_to_array(ctx.table)
    tables = find_tables(ctx)
    matching_tables = []

    if nth is not None:
        if assert_func(tables[nth], expected):
            matching_tables.append(tables[nth])

    else:
        for table in tables:
            if assert_func(table, expected):
                matching_tables.append(table)

    # If none of the tables match the pattern, then return
    if len(matching_tables) == 0:
        return

    report_found_undesired_table(matching_tables, tables)


for thing, check_func in {
    "is": check_table_equals_table,
    "matches": check_table_matches_table,
    "contains": check_table_contains_table,
    "contains rows matching": check_table_contains_matching_rows_in_table,
}.items():

    @step(f"I should see a table that {thing} the following")
    def should_see_the_table(ctx, check_func=check_func):
        find_table(ctx, check_func)

    @step(f"I should not see a table that {thing} the following")
    def should_not_see_the_table(ctx, check_func=check_func):
        do_not_find_table(ctx, check_func)

    @step(f"I wait to see a table that {thing} the following")
    def wait_to_see_the_table(ctx, check_func=check_func):
        retry(find_table)(ctx, check_func)

    @step(f"I wait to not see a table that {thing} the following")
    def wait_to_not_see_the_table(ctx, check_func=check_func):
        retry(do_not_find_table)(ctx, check_func)

    @step(
        f'I wait up to "{{seconds}}" seconds to see a table that {thing} the following'
    )
    def wait_up_to_seconds_to_see_the_table(
        ctx, seconds, check_func=check_func
    ):
        seconds = float(seconds)
        retry(find_table, wait_up_to_s=seconds)(ctx, check_func)

    @step(
        f'I wait up to "{{seconds}}" seconds to not see a table that {thing} the following'
    )
    def wait_up_to_seconds_to_not_see_the_table(
        ctx, seconds, check_func=check_func
    ):
        seconds = float(seconds)
        retry(do_not_find_table, wait_up_to_s=seconds)(ctx, check_func)

    @step(f'I should see the "{{nth}}" table {thing} the following')
    def should_see_the_nth_table(ctx, nth, check_func=check_func):
        find_table(ctx, check_func, nth=nth)

    @step(f'I wait to see the "{{nth}}" table {thing} the following')
    def wait_to_see_the_nth_table(ctx, nth, check_func=check_func):
        retry(find_table)(ctx, check_func, nth=nth)

    @step(
        f'I wait up to "{{seconds}}" seconds to see the "{{nth}}" table {thing} the following'
    )
    def wait_up_to_seconds_to_see_the_nth_table(
        ctx, seconds, nth, check_func=check_func
    ):
        seconds = float(seconds)
        retry(find_table, wait_up_to_s=seconds)(ctx, check_func, nth=nth)


def find_table_header(ctx, name, index=0):
    """
    find a table header with the provided name
    """
    ctx.check_browser_initialized()
    return fuzzy.find(ctx.browser, name, ["th"], index=index)


def click_table_header(ctx, header):
    """
    internal method used to simply click a table header element
    """
    ctx.check_browser_initialized()
    ctx.browser.click(header)


helpers.define_action_on_thing_with_name_steps(
    "table header",
    "click",
    find_table_header,
    click_table_header,
    with_nth=True,
)


def get_table_cell_value(ctx, table, row, column, variable_name):
    tables = find_tables(ctx)

    try:
        cell_value = tables[table][row][column]
    except IndexError:
        raise RuntimeError(
            f"Cannot find table:{table + 1},row:{row + 1},column:{column + 1}. Please check your table data."
        )
    config.CONFIG[variable_name] = cell_value


@step(
    'I save "{table:nth}" table, "{row:nth}" row, "{column:nth}" column value to a variable "{variable_name}"'
)
def step_get_table_cell_value(ctx, table, row, column, variable_name):
    get_table_cell_value(ctx, table, row, column, variable_name)


@step(
    'I wait to save "{table:nth}" table, "{row:nth}" row, "{column:nth}" column value to a variable "{variable_name}"'
)
def wait_to_get_table_cell_value(ctx, table, row, column, variable_name):
    retry(get_table_cell_value)(ctx, table, row, column, variable_name)


def find_table_element(ctx, nth=1):
    """
    Return the nth table as a WebElement

    parameters:
      ctx(object): behave context object used to share data between steps
      nth(int): specifies the exact table within the list of available tables to match against.
                Defaults to 1st table.

    returns:
      A selenium WebElement associated with the table that was specified
    """
    ctx.check_browser_initialized()

    try:
        return ctx.browser.css_find_elements("table")[nth]
    except IndexError:
        raise RuntimeError(
            f"Cannot find table:{nth + 1}. Please check your table data."
        )


def click_table_cell(ctx, row, column, table):
    """
    Clicks the cell corresponding to the given row and column

    parameters:
      ctx(object): behave context object used to share data between steps
      row(int): the row of the table to click
      column(int): the column of the table to click
      table(int): specifies the exact table within the list of available tables to match against.
    """
    table_element = find_table_element(ctx, table)

    try:
        row = table_element.find_elements(By.CSS_SELECTOR, "tbody tr")[row]
        cell = row.find_elements(By.CSS_SELECTOR, "td")[column]
    except IndexError:
        raise RuntimeError(
            f"Cannot find table:{table + 1},row:{row + 1},column:{column + 1}. Please check your table data."
        )
    ctx.browser.click(cell)


@step('I wait to click the "{row:nth}" row in the "{table:nth}" table')
def wait_click_table_row(ctx, row, table):
    """
    Add 1 to the row number if the table has a header row.

    Note: Firefox is unable to click directly on a row <tr> if it has child columns <td>.
    In order to workaround this, the step just clicks the first column <td> of the row <tr>.
    Bug: https://bugzilla.mozilla.org/show_bug.cgi?id=1448825
    """
    retry(click_table_cell)(ctx, row, 1, table)


@step(
    'I wait to click the cell corresponding to the "{row:nth}" row and "{column:nth}" column in the "{table:nth}" table'
)
def wait_click_table_cell(ctx, row, column, table):
    """
    Add 1 to the row number if the table has a header row.
    """
    retry(click_table_cell)(ctx, row, column, table)


@step(
    'I wait to click the "{column:nth}" column within a row that contains the text "{match_text}" in the "{table:nth}" table'
)
def wait_click_table_cell_matching_text(ctx, column, match_text, table):
    def click_table_cell_matching_text(ctx, column, match_text, table):
        table_element = find_table_element(ctx, table)

        try:
            row = table_element.find_elements(
                By.XPATH, f'//td[.="{match_text}"]/parent::tr'
            )
            if len(row) > 1:
                logger.warning(
                    f'Found {len(row)} rows with matching text "{match_text}", using the first row.'
                )
            cell = row[0].find_elements(By.CSS_SELECTOR, "td")[column]
        except IndexError:
            raise RuntimeError(
                f"Cannot find table:{table + 1},column:{column + 1},text:{match_text}. Please check your table data."
            )

        ctx.browser.click(cell)

    retry(click_table_cell_matching_text)(ctx, column, match_text, table)


@step('I wait to see there are "{row_count}" rows in the "{table:nth}" table')
def wait_table_row_count(ctx, row_count, table):
    """
    Add 1 to the row number if the table has a header row.
    """

    def find_table_row_count(ctx, row_count, table):
        table_element = find_table_element(ctx, table)
        table_rows = len(table_element.find_elements(By.CSS_SELECTOR, "tr"))

        if int(row_count) == table_rows:
            return
        else:
            raise RuntimeError(
                f"Unable to find {row_count} rows in table {table + 1}. Please check your table data."
            )

    retry(find_table_row_count)(ctx, row_count, table)



import html
import re

from behave.formatter.ansi_escapes import colors, escapes

ESC_SEQ = r"\x1b["
TRANSLATION = (
    {colors["grey"]: '<span style="color: grey;">\n'}
    | {  # grey indicates a new step, and hence add a line break
        v: f'<span style="color: {k};">'
        for k, v in colors.items()
        if k != "grey"
    }
    | {
        "\n" + escapes["reset"] + "\n": "</span>\n",
        escapes["reset"]: "</span>",
        escapes["up"] + "\n": "",  # remove a line break with cursor up
        ESC_SEQ + "46": "",  # ignore DELETE (num keypad)
        ESC_SEQ + "48": "",  # ignore INSERT (num keypad)
        ESC_SEQ + "49": "",  # ignore END (num keypad)
        ESC_SEQ + "50": "",  # ignore DOWN ARROW (num keypad)
        ESC_SEQ + "51": "",  # ignore PAGE DOWN (num keypad)
        ESC_SEQ + "52": "",  # ignore LEFT ARROW (num keypad)
        ESC_SEQ + "54": "",  # ignore RIGHT ARROW (num keypad)
        ESC_SEQ + "55": "",  # ignore HOME (num keypad)
        ESC_SEQ + "56": "",  # ignore UP ARROW (num keypad)
        ESC_SEQ + "57": "",  # ignore PAGE UP (num keypad)
        ESC_SEQ + "1A": "",  # ignore move cursor up one line
    }
)
RE_TO_HTML = re.compile("|".join(map(re.escape, TRANSLATION)))

RE_TO_REMOVE = re.compile(
    r"\x1b\[(0;)?[0-9A-F]{1,2}[mA]:?"
)  # detect hex values, not just decimal digits


def remove_ansi(input: str) -> str:
    return RE_TO_REMOVE.sub("", input)


def parse_log_to_html(input: str) -> str:
    """
    Parse an ansi color log to html
    """
    html_start = "<!DOCTYPE html>\n<html>\n"
    html_end = "</html>\n"
    head = '<head>\n<meta http-equiv="Content-Type" content="text/html; charset=utf-8">\n</head>\n'
    body_start = '<body style="color: white; background-color: #100c08;">'  # use dark bg since colors are from behave
    body_end = "</body>\n"
    result = f"{html_start}{head}{body_start}<pre>\n{RE_TO_HTML.sub(lambda match: TRANSLATION[match.group(0)], html.escape(input, quote=False))}\n</pre>{body_end}{html_end}"
    if ESC_SEQ in result:
        lines = "\n".join([x for x in result.split("\n") if ESC_SEQ in x])

        print(
            f"Detected unmapped ansi escape code!:\n{lines}"
        )  # use print instead of logger to avoid circular import

    return result



import glob
import importlib
import os
import re

import yaml

from cucu import logger
from cucu.cli.steps import load_cucu_steps
from cucu.config import CONFIG


def load_lint_rules(rules, filepath):
    """
    load all of the lint rules from the filepath provided

    Returns:
        hashmap of of lint rules
    """
    filepath = os.path.abspath(filepath)

    if os.path.isdir(filepath):
        basepath = os.path.join(filepath, "**/*.yaml")
        lint_rule_filepaths = glob.iglob(basepath, recursive=True)

    else:
        lint_rule_filepaths = [filepath]

    for lint_rule_filepath in lint_rule_filepaths:
        logger.debug(f"loading lint rules from {lint_rule_filepath}")

        with open(lint_rule_filepath, "r", encoding="utf8") as _input:
            rules_loaded = yaml.safe_load(_input.read())

        for rule_name, rule in rules_loaded.items():
            if rule_name in rules:
                raise RuntimeError(
                    f"found duplicate rule names {rule_name}, please correct one of the locations."
                )

            rules[rule_name] = rule

    return rules


def parse_matcher(name, rule_name, rule, line, state):
    """
    parses the "matcher" from the rule provided and then returns the tuple:
    (matched, extra_matcher_message) where matched is a boolean indicating that
    the rule name matcher name and rule matched on the line provided and the
    extra_matcher_message is used when reporting the linting failure upstream.

    name(string): name of the line to match on, ie current_line, previous_line,
                  next_line are currently supported
    rule_name(string): the actual name of the rule from the rules file, such as:
                       when_keyword_indented_correctly
    rule(dict): the rule dictionary object which contains the matcher, fix, etc.
    line(string): actual line to parse the matcher against
    state(dict): state object passed in which contains a few things such as:
                 current_feature_filepath, current_feature_name, etc.

    returns: a tuple where the first element is True if the matcher matches the
             specified line. False if the matcher simply doesn't apply to this
             line. The second part of the tuple is an string used to augment
             the lint violation if the remainder of the matchers in a rule
             matched their specific line (current_line, previous_line, etc).
    """
    if name not in rule:
        # when the name provided isn't in the rule then we simply have a no-op
        # where this specific matcher name wasn't even used.
        return (True, "")

    if "match" in rule[name]:
        if line is None:
            return (False, "")

        match = re.match(rule[name]["match"], line)

        if match is None:
            # no match on the specified rule means there's not violation to
            # report
            return (False, "")

        cwd = f"{os.getcwd()}/"

        # unique across all feature files
        if "unique_per_all_features" in rule[name]:
            value = match.groups()[0]
            feature_filepath = state["current_feature_filepath"]
            # make the path relative to the current working directory
            feature_filepath = feature_filepath.replace(cwd, "")

            if rule_name not in state["unique_per_all_features"]:
                state["unique_per_all_features"][rule_name] = {}

            if value in state["unique_per_all_features"][rule_name]:
                # we have another feature which already has this value in use.
                other_filepath = state["unique_per_all_features"][rule_name][
                    value
                ]
                # make the path relative to the current working directory
                other_filepath = other_filepath.replace(cwd, "")

                if other_filepath != feature_filepath:
                    return (
                        True,
                        f', "{value}" also used in "{other_filepath}"',
                    )

            state["unique_per_all_features"][rule_name][value] = (
                feature_filepath
            )
            return (False, "")

        # unique across all scenarios across all features
        if "unique_per_all_scenarios" in rule[name]:
            value = match.groups()[0]
            feature_filepath = state["current_feature_filepath"]
            # make the path relative to the current working directory
            feature_filepath = feature_filepath.replace(cwd, "")
            scenario_name = state["current_scenario_name"]

            if rule_name not in state["unique_per_all_scenarios"]:
                state["unique_per_all_scenarios"][rule_name] = {}

            if value in state["unique_per_all_scenarios"][rule_name]:
                # we have another scenario which already has this value in use.
                other_file_path, other_line_number, other_scenario_name = (
                    state["unique_per_all_scenarios"][rule_name][value]
                )

                return (
                    True,
                    f', "{value}" also used in "{other_file_path}:{other_line_number}" Scenario: "{other_scenario_name}"',
                )

            state["unique_per_all_scenarios"][rule_name][value] = [
                feature_filepath,
                state["current_line_number"],
                scenario_name,
            ]

            return (False, "")

        return (True, "")

    raise RuntimeError(f"unsupported matcher for {name}")


def lint_line(state, rules, steps, line_number, lines, filepath):
    """ """
    if line_number >= 1:
        previous_line = lines[line_number - 1]

    else:
        previous_line = None

    current_line = lines[line_number]

    if line_number + 1 < len(lines):
        next_line = lines[line_number + 1]
    else:
        next_line = None

    logger.debug(f'linting line "{current_line}"')

    violations = []
    for rule_name in rules.keys():
        logger.debug(f' * checking against rule "{rule_name}"')
        rule = rules[rule_name]

        # skip paths that match the exclude regex
        if "exclude" in rule and re.match(rule["exclude"], filepath):
            continue

        (current_matcher, current_message) = parse_matcher(
            "current_line",
            rule_name,
            rule,
            current_line,
            state,
        )
        (previous_matcher, previous_message) = parse_matcher(
            "previous_line",
            rule_name,
            rule,
            previous_line,
            state,
        )
        (next_matcher, next_message) = parse_matcher(
            "next_line",
            rule_name,
            rule,
            next_line,
            state,
        )

        logger.debug(
            f"previous matcher {previous_matcher} current matcher {current_matcher} next matcher {next_matcher}"
        )

        if current_matcher and previous_matcher and next_matcher:
            type = rule["type"][0].upper()
            message = rule["message"]

            if "fix" in rule:
                fix = rule["fix"]
            else:
                fix = None

            violations.append(
                {
                    "location": {
                        "filepath": os.path.relpath(filepath),
                        "line": line_number,
                    },
                    "type": type,
                    "message": f"{message}{previous_message}{current_message}{next_message}",
                    "fix": fix,
                }
            )

    # find any undefined steps and mark them as an unfixable violation
    current_line = current_line.strip()
    undefined_steps = [
        {
            "location": {
                "filepath": os.path.relpath(filepath),
                "line": line_number,
            },
            "type": "error",
            "message": f'undefined step "{step_name}"',
            "fix": None,
        }
        for step_name in steps
        # step with no location/type/etc is an undefined step
        if steps[step_name] is None and current_line.find(step_name) != -1
    ]
    violations.extend(undefined_steps)

    return violations


def fix(violations):
    """
    fix the violations found in a set of violations relating to a single
    feature file.
    """
    if not violations:
        return

    deletions = []
    filepath = violations[0]["location"]["filepath"]
    lines = open(filepath, "r").read().split("\n")

    for violation in violations:
        if violation["fix"] is None:
            violation["fixed"] = False

        else:
            line_number = violation["location"]["line"]
            line_to_fix = lines[line_number]

            if "delete" in violation["fix"]:
                # store the deletions to do at the end
                deletions.append(violation)

            elif "match" in violation["fix"]:
                match = violation["fix"]["match"]
                replace = violation["fix"]["replace"]
                fixed_line = re.sub(match, replace, line_to_fix)
                lines[line_number] = fixed_line
                violation["fixed"] = True

            else:
                raise RuntimeError(f"unknown fix type in {violation}")

    # sort the deletions from bottom to the top of the file and then perform
    # the deletions
    deletions.sort(key=lambda x: x["location"]["line"], reverse=True)
    for violation in deletions:
        del lines[violation["location"]["line"]]
        violation["fixed"] = True

    with open(filepath, "w") as output:
        output.write("\n".join(lines))

    return violations


def load_builtin_lint_rules(rules):
    """
    load internal builtin lint rules and used primarily for unit testing
    """
    cucu_path = os.path.dirname(importlib.util.find_spec("cucu").origin)
    lint_rules_path = os.path.join(cucu_path, "lint", "rules")
    load_lint_rules(rules, lint_rules_path)


def lint(filepath):
    """
    lint the filepath provided which could be a base directory where many
    feature files exists and we must traverse recursively or a specific file.

    Params:
        filepath(string): path to a directory or feature file to lint

    Returns:
        a generator of violations per file, so each value yielded to the
        generator is a list of violations within the same file
    """
    rules = {}

    logger.debug(f"linting {filepath}")

    # load the base lint rules
    load_builtin_lint_rules(rules)

    if CONFIG["CUCU_LINT_RULES_PATH"]:
        # load any other rules paths linked via CUCU_LINT_RULES_PATH variable
        lint_rule_paths = CONFIG["CUCU_LINT_RULES_PATH"].split(",")

        for lint_rule_path in lint_rule_paths:
            logger.debug(f"loading custom rules from: {lint_rule_path}")
            load_lint_rules(rules, lint_rule_path)

    steps, steps_error = load_cucu_steps(filepath=filepath)

    # state object used to carry state from the top level linting function down
    # to the functions handling the lint rules and reporting on lint failures
    state = {
        "unique_per_all_features": {},
        "unique_per_all_scenarios": {},
    }

    if steps_error:
        yield [
            {
                "type": "steps_error",
                "message": steps_error,
            }
        ]

    filepath = os.path.abspath(filepath)

    if os.path.isdir(filepath):
        basepath = os.path.join(filepath, "**/*.feature")
        # XXX: for now sorted by name... we could expose some options for other
        #      sorting orders if it makes sense
        feature_filepaths = sorted(glob.iglob(basepath, recursive=True))

    else:
        feature_filepaths = [filepath]

    for feature_filepath in feature_filepaths:
        state["current_feature_filepath"] = feature_filepath

        lines = open(feature_filepath).read().split("\n")
        line_number = 0

        violations = []
        in_docstring = {
            '"""': False,
            "'''": False,
        }

        for line in lines:
            state["current_line_number"] = line_number

            feature_match = re.match(".*Feature: (.*)", line)
            if feature_match is not None:
                state["current_feature_name"] = feature_match.group(1)

            scenario_match = re.match("  Scenario: (.*)", line)

            if scenario_match is not None:
                state["current_scenario_name"] = scenario_match.group(1)
            else:
                state["current_scenario_name"] = ""

            # maintain state of if we're inside a docstring and if we are then
            # do not apply any linting rules as its a freeform space for text
            if line.strip() == '"""':
                in_docstring['"""'] = not in_docstring['"""']

            if line.strip() == "'''":
                in_docstring["'''"] = not in_docstring["'''"]

            if not (in_docstring['"""'] or in_docstring["'''"]):
                for violation in lint_line(
                    state, rules, steps, line_number, lines, feature_filepath
                ):
                    violations.append(violation)

            line_number += 1

        yield violations



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import traceback

from behave.formatter.ansi_escapes import colors, escapes, up
from behave.formatter.base import Formatter
from behave.model_core import Status
from behave.model_describe import ModelPrinter
from behave.textutil import make_indentation

from cucu.config import CONFIG


class CucuFormatter(Formatter):
    """ """

    name = "cucu"
    description = "cucu specific formatter to make the logs more readable"

    DEFAULT_INDENT_SIZE = 2
    SUBSTEP_PREFIX = "  ⤷"

    def __init__(self, stream_opener, config, **kwargs):
        super(CucuFormatter, self).__init__(stream_opener, config)
        self.current_scenario = None
        self.steps = []
        self.match_step_index = 0
        self.show_timings = config.show_timings
        self.indent_size = self.DEFAULT_INDENT_SIZE
        # -- ENSURE: Output stream is open.
        self.stream = self.open()
        self.printer = ModelPrinter(self.stream)
        # -- LAZY-EVALUATE:
        self._multiline_indentation = None

        color_output = CONFIG["CUCU_COLOR_OUTPUT"]
        self.monochrome = color_output != "true"

    @property
    def multiline_indentation(self):
        if self._multiline_indentation is None:
            # align keywords
            offset = 2
            indentation = make_indentation(3 * self.indent_size + offset)
            self._multiline_indentation = indentation

        return self._multiline_indentation

    def write_tags(self, tags, indent=None, for_feature=False):
        """
        writes the tags for a scenario or feature and takes extra care to be
        sure to indent corectly for scenarios when `indent` is specified. We
        also handle putting new lines in the right locations if its a scenario
        or feature based on `for_feature` boolean flag.
        """
        indent = indent or ""
        if tags:
            text = " @".join(tags)
            line = self.colorize(f"{indent}@{text}", "cyan")
            if for_feature:
                self.stream.write(f"{line}\n")
            else:
                self.stream.write(f"\n{line}")

    # -- IMPLEMENT-INTERFACE FOR: Formatter
    def feature(self, feature):
        self.write_tags(feature.tags, for_feature=True)
        text = f"{self.colorize(feature.keyword, 'magenta')}: {feature.name}\n"
        self.stream.write(text)

    def colorize(self, text, color):
        if self.monochrome:
            return text
        else:
            return colors[color] + text + escapes["reset"]

    def background(self, background):
        if not self.monochrome:
            indent = make_indentation(self.indent_size)
            keyword = self.colorize(background.keyword, "magenta")
            text = f"\n{indent}{keyword}: {background.name}\n"
            self.stream.write(text)

    def scenario(self, scenario):
        self.current_scenario = scenario

        indent = make_indentation(self.indent_size)
        text = "\n%s%s: %s\n" % (
            indent,
            self.colorize(scenario.keyword, "magenta"),
            scenario.name,
        )
        self.write_tags(scenario.tags, indent)
        self.stream.write(text)
        self.steps = []
        self.match_step_index = 0

    def step(self, step):
        self.insert_step(step)

    def insert_step(self, step, index=-1):
        # used to determine how to better handle console output
        step.has_substeps = getattr(step, "has_substeps", False)
        step.is_substep = getattr(step, "is_substep", False)

        if index == -1:
            self.steps.append(step)
        else:
            self.steps.insert(index, step)

    def match(self, match):
        if self.monochrome:
            return

        # we'll write a step line in grey and not hit the carriage return
        step = self.steps[self.match_step_index]
        self.match_step_index += 1

        indent = make_indentation(2 * self.indent_size)
        keyword = step.keyword.rjust(5)

        prefix = ""
        if step.is_substep:
            prefix = self.SUBSTEP_PREFIX

        text = self.colorize(
            f"{indent}{prefix}{keyword} {step.name}\n", "grey"
        )
        self.stream.write(text)
        self.stream.flush()

    def calculate_max_line_length(self):
        line_lengths = [
            len(f"{step.keyword.rjust(5)} {step.name}") for step in self.steps
        ]

        return max(line_lengths) + 4

    def result(self, step):
        indent = make_indentation(2 * self.indent_size)
        keyword = step.keyword.rjust(5)

        if not (
            self.monochrome
            or step.has_substeps
            or CONFIG["__CUCU_WROTE_TO_OUTPUT"]
        ):
            self.stream.write(up(1))

        prefix = ""
        if step.is_substep:
            prefix = self.SUBSTEP_PREFIX

        if step.status == Status.passed:
            keyword = self.colorize(keyword, "green")
            text = f"{indent}{prefix}{keyword} {step.name}"
        elif step.status == Status.failed:
            text = self.colorize(
                f"{indent}{prefix}{keyword} {step.name}", "red"
            )
        elif step.status == Status.undefined:
            text = self.colorize(
                f"{indent}{prefix}{keyword} {step.name}", "yellow"
            )
        elif step.status == Status.skipped:
            text = self.colorize(
                f"{indent}{prefix}{keyword} {step.name}\n", "cyan"
            )
        elif step.status == Status.untested:
            text = self.colorize(
                f"{indent}{prefix}{keyword} {step.name}\n", "cyan"
            )
        else:
            text = self.colorize(
                f"{indent}{prefix}{keyword} {step.name}\n", "yellow"
            )

        if self.monochrome:
            self.stream.write(f"{text}")
        else:
            self.stream.write(f"\r{text}")

        if step.status in (Status.passed, Status.failed):
            max_line_length = self.calculate_max_line_length()
            status_text = ""
            if self.show_timings:
                start = step.start_at
                duration = f"{step.duration:.3f}"
                status_text += f" # started at {start} took {duration}s"

            current_step_text = f"{step.keyword.rjust(5)} {step.name}"
            status_text_padding = (
                max_line_length - len(current_step_text) - len(prefix)
            )
            status_text = f"{' ' * status_text_padding}{status_text}"
            status_text = self.colorize(status_text, "yellow")

            self.stream.write(f"{status_text}\n")
            if step.error_message:
                self.stream.write(f"{step.error_message}\n")

        if step.text:
            self.doc_string(step.text)

        if step.table:
            self.table(getattr(step.table, "original", step.table))

        if step.status in (Status.passed, Status.failed):
            # print the variable values in step name, multiline/table arguments
            # as a comment to ease debugging
            step_variables = CONFIG.expand(step.name)

            if step.text:
                step_variables.update(CONFIG.expand(step.text))

            if step.table:
                for row in step.table.original.rows:
                    for value in row:
                        step_variables.update(CONFIG.expand(value))

            if step_variables:
                expanded = " ".join(
                    [
                        f'{key}="{value}"'
                        for (key, value) in step_variables.items()
                    ]
                )

                padding = f"    {' ' * (len('Given') - len(step.keyword))}"
                variable_line = f"{padding}# {expanded}\n"
                # hide secrets before we do anything to add color which could
                # modify the output and result in not being able to correctly
                # parse
                # TODO: I'd like to move this out of here as we should be able
                #       to intercept all of the stdout/stderr writes but seems
                #       behaves underlying self.stream here is getting around
                #       that by accessing stdout/stderr another way.
                variable_line = CONFIG.hide_secrets(variable_line)
                colored_variable_line = self.colorize(variable_line, "blue")
                self.stream.write(colored_variable_line)
                self.stream.flush()

        self.previous_step = step
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = False

    def eof(self):
        self.stream.write("\n")

        if self.current_scenario:
            if self.current_scenario.status.name == "failed":
                # we need to record the error_message and exc_traceback in the
                # last executed step and mark it as failed so the reporting can
                # show the result correctly
                error_message = traceback.format_tb(
                    self.current_scenario.exc_traceback
                )
                self.stream.write("\n".join(error_message))

    # -- MORE: Formatter helpers
    def doc_string(self, doc_string):
        self.printer.print_docstring(doc_string, self.multiline_indentation)

    def table(self, table):
        self.printer.print_table(table, self.multiline_indentation)



"""
Database creation and management utilities for cucu.
"""

import logging
import sqlite3
import sys
from pathlib import Path

from peewee import (
    BooleanField,
    DateTimeField,
    FloatField,
    ForeignKeyField,
    IntegerField,
    Model,
    TextField,
)
from playhouse.sqlite_ext import JSONField, SqliteExtDatabase
from tenacity import RetryError

from cucu import logger as cucu_logger
from cucu.config import CONFIG
from cucu.utils import get_iso_timestamp_with_ms, parse_iso_timestamp

db_filepath = CONFIG["RUN_DB_PATH"]
db = SqliteExtDatabase(db_filepath)

logger = logging.getLogger("peewee")
logger.setLevel(logging.WARNING)  # Only show warnings and errors


class BaseModel(Model):
    class Meta:
        database = db


class cucu_run(BaseModel):
    cucu_run_id = TextField(primary_key=True)
    full_arguments = JSONField()
    filepath = TextField()
    start_at = DateTimeField()
    end_at = DateTimeField(null=True)
    db_path = TextField(null=True)
    run_info = JSONField(null=True)


class worker(BaseModel):
    worker_run_id = TextField(primary_key=True)
    cucu_run = ForeignKeyField(
        cucu_run,
        backref="workers",
        column_name="cucu_run_id",
        null=True,
    )
    parent_id = ForeignKeyField(
        "self",
        field="worker_run_id",
        backref="child_workers",
        column_name="parent_run_id",
        null=True,
    )
    start_at = DateTimeField()
    end_at = DateTimeField(null=True)
    custom_data = JSONField(null=True)


class feature(BaseModel):
    feature_run_id = TextField(primary_key=True)
    worker = ForeignKeyField(
        worker,
        backref="features",
        column_name="worker_run_id",
    )
    name = TextField()
    status = TextField(null=True)
    start_at = DateTimeField()
    end_at = DateTimeField(null=True)
    custom_data = JSONField(null=True)
    tags = JSONField()
    filename = TextField()
    description = TextField()
    behave_filepath = TextField()


class scenario(BaseModel):
    scenario_run_id = TextField(primary_key=True)
    feature = ForeignKeyField(
        feature,
        backref="scenarios",
        column_name="feature_run_id",
    )
    name = TextField()
    seq = IntegerField(null=True)
    status = TextField(null=True)
    duration = FloatField(null=True)
    start_at = DateTimeField(null=True)
    end_at = DateTimeField(null=True)
    tags = JSONField()
    custom_data = JSONField(null=True)
    browser_info = JSONField(null=True)
    cucu_config = JSONField(null=True)
    line_number = IntegerField()
    log_files = JSONField(null=True)


class step(BaseModel):
    step_run_id = TextField(primary_key=True)
    scenario = ForeignKeyField(
        scenario,
        backref="steps",
        column_name="scenario_run_id",
    )
    section_level = IntegerField(null=True)
    seq = IntegerField()
    parent_seq = IntegerField(null=True)
    is_substep = BooleanField(null=True)  # info available after step ends
    has_substeps = BooleanField()
    keyword = TextField()
    name = TextField()
    status = TextField(null=True)
    duration = FloatField(null=True)
    start_at = DateTimeField(null=True)
    end_at = DateTimeField(null=True)
    stdout = JSONField()
    stderr = JSONField()
    error_message = JSONField(null=True)
    exception = JSONField(null=True)
    debug_output = JSONField()
    browser_info = JSONField()
    text = JSONField(null=True)
    table_data = JSONField(null=True)
    location = TextField()
    browser_logs = JSONField()
    screenshots = JSONField()
    image_dir = TextField(null=True)


def record_cucu_run():
    filepath = CONFIG["CUCU_FILEPATH"]
    cucu_run_id_val = CONFIG["CUCU_RUN_ID"]
    worker_run_id = CONFIG["WORKER_RUN_ID"]

    db.connect(reuse_if_open=True)
    cucu_run.create(
        cucu_run_id=cucu_run_id_val,
        full_arguments=sys.argv,
        filepath=filepath,
        start_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
    )

    parent_id = (
        CONFIG.get("WORKER_PARENT_ID")
        if CONFIG.get("WORKER_PARENT_ID") != worker_run_id
        else None
    )
    worker.create(
        worker_run_id=worker_run_id,
        cucu_run_id=cucu_run_id_val,
        parent_id=parent_id,
        start_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
    )

    return str(db_filepath)


def record_feature(feature_obj):
    db.connect(reuse_if_open=True)
    feature.create(
        feature_run_id=feature_obj.feature_run_id,
        worker=CONFIG["WORKER_RUN_ID"],
        name=feature_obj.name,
        filename=feature_obj.filename,
        description="\n".join(feature_obj.description)
        if isinstance(feature_obj.description, list)
        else str(feature_obj.description),
        tags=feature_obj.tags,
        start_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
        behave_filepath=CONFIG["BEHAVE_FILEPATH"],
    )


def record_scenario(scenario_obj):
    db.connect(reuse_if_open=True)
    scenario.create(
        scenario_run_id=scenario_obj.scenario_run_id,
        feature_run_id=scenario_obj.feature.feature_run_id,
        name=scenario_obj.name,
        line_number=scenario_obj.line,
        seq=scenario_obj.seq,
        tags=scenario_obj.tags,
        start_at=parse_iso_timestamp(getattr(scenario_obj, "start_at", None)),
    )


def start_step_record(step_obj, scenario_run_id):
    db.connect(reuse_if_open=True)

    table = None
    if step_obj.table:
        table = {
            "headings": step_obj.table.headings,
            "rows": [list(row) for row in step_obj.table.rows],
        }

    step.create(
        step_run_id=step_obj.step_run_id,
        scenario_run_id=scenario_run_id,
        seq=getattr(step_obj, "seq", -1),
        keyword=step_obj.keyword,
        name=step_obj.name,
        status=step_obj.status.name,
        text=step_obj.text.splitlines() if step_obj.text else [],
        table_data=table,
        location=str(step_obj.location),
        is_substep=getattr(step_obj, "is_substep", False),
        has_substeps=getattr(step_obj, "has_substeps", False),
        section_level=getattr(step_obj, "section_level", None),
        browser_info="",
        browser_logs=[],
        error_message=[],
        debug_logs=[],
        debug_output=[],
        stderr=[],
        stdout=[],
        screenshots=[],
    )


def finish_step_record(step_obj, duration):
    db.connect(reuse_if_open=True)

    error_message = []
    exception = []
    if step.error_message and step_obj.status.name == "failed":
        error_message = CONFIG.hide_secrets(
            step_obj.error_message
        ).splitlines()

        if error := step_obj.exception:
            if isinstance(error, RetryError):
                error = error.last_attempt.exception()

            if len(error.args) > 0 and isinstance(error.args[0], str):
                error_class_name = error.__class__.__name__
                redacted_error_msg = CONFIG.hide_secrets(error.args[0])
                error_lines = redacted_error_msg.splitlines()
                error_lines[0] = f"{error_class_name}: {error_lines[0]}"
            else:
                error_lines = [repr(error)]

            exception = error_lines

    step.update(
        browser_info=getattr(step_obj, "browser_info", {}),
        browser_logs=getattr(step_obj, "browser_logs", []),
        debug_output=getattr(step_obj, "debug_output", []),
        duration=duration,
        end_at=getattr(step_obj, "end_at", None),
        error_message=error_message,
        exception=exception,
        has_substeps=getattr(step_obj, "has_substeps", False),
        parent_seq=getattr(step_obj, "parent_seq", None),
        screenshots=getattr(step_obj, "screenshots", []),
        section_level=getattr(step_obj, "section_level", None),
        seq=step_obj.seq,
        start_at=parse_iso_timestamp(getattr(step_obj, "start_at", None)),
        status=step_obj.status.name,
        stderr=getattr(step_obj, "stderr", []),
        stdout=getattr(step_obj, "stdout", []),
        image_dir=getattr(step_obj, "step_image_dir", None),
    ).where(step.step_run_id == step_obj.step_run_id).execute()


def finish_scenario_record(scenario_obj):
    db.connect(reuse_if_open=True)
    start_at = parse_iso_timestamp(getattr(scenario_obj, "start_at", None))
    end_at = parse_iso_timestamp(getattr(scenario_obj, "end_at", None))
    if start_at and end_at:
        duration = (end_at - start_at).total_seconds()
    else:
        duration = None

    scenario_logs_dir = CONFIG.get("SCENARIO_LOGS_DIR")
    if not scenario_logs_dir or not Path(scenario_logs_dir).exists():
        log_files_json = "[]"
    else:
        logs_path = Path(scenario_logs_dir)
        log_files = [
            str(file.relative_to(logs_path))
            for file in logs_path.rglob("*")
            if file.is_file()
        ]
        log_files_json = sorted(log_files)

    if scenario_obj.hook_failed:
        status = "errored"
    else:
        status = scenario_obj.status.name

    scenario.update(
        status=status,
        duration=duration,
        end_at=end_at,
        log_files=log_files_json,
        cucu_config=getattr(scenario_obj, "cucu_config_json", dict()),
        browser_info=getattr(scenario_obj, "browser_info", dict()),
        custom_data=getattr(scenario_obj, "custom_data", dict()),
    ).where(scenario.scenario_run_id == scenario_obj.scenario_run_id).execute()


def finish_feature_record(feature_obj):
    db.connect(reuse_if_open=True)
    feature.update(
        status=feature_obj.status.name,
        end_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
        custom_data=feature_obj.custom_data,
    ).where(feature.feature_run_id == feature_obj.feature_run_id).execute()


def finish_worker_record(custom_data=None, worker_run_id=None):
    db.connect(reuse_if_open=True)
    target_worker_run_id = worker_run_id or CONFIG["WORKER_RUN_ID"]
    worker.update(
        end_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
        custom_data=custom_data,
    ).where(worker.worker_run_id == target_worker_run_id).execute()


def finish_cucu_run_record():
    db.connect(reuse_if_open=True)
    cucu_run.update(
        end_at=parse_iso_timestamp(get_iso_timestamp_with_ms()),
    ).where(cucu_run.cucu_run_id == CONFIG["CUCU_RUN_ID"]).execute()


def close_db():
    if not db.is_closed():
        db.close()


def create_database_file(db_filepath):
    db.init(db_filepath)
    db.connect(reuse_if_open=True)
    db.create_tables([cucu_run, worker, feature, scenario, step])
    db.execute_sql("""
            CREATE VIEW IF NOT EXISTS flat_all AS
            WITH scenario_with_steps AS (
                SELECT
                    *,
                    COUNT(st.step_run_id) AS steps
                FROM scenario s
                LEFT JOIN step st ON s.scenario_run_id = st.scenario_run_id
                GROUP BY s.scenario_run_id
            )
            SELECT
                COUNT(DISTINCT s.feature_run_id) AS features,
                COUNT(s.scenario_run_id) AS scenarios,
                SUM(CASE WHEN s.status = 'passed' THEN 1 ELSE 0 END) AS passed,
                SUM(CASE WHEN s.status = 'failed' THEN 1 ELSE 0 END) AS failed,
                SUM(CASE WHEN s.status = 'skipped' THEN 1 ELSE 0 END) AS skipped,
                SUM(CASE WHEN s.status = 'errored' THEN 1 ELSE 0 END) AS errored,
                SUM(s.duration) AS duration,
                SUM(s.steps) AS steps
            FROM scenario_with_steps s
        """)
    db.execute_sql("""
            CREATE VIEW IF NOT EXISTS flat_feature AS
            WITH feature_first_level AS (
                SELECT
                    w.cucu_run_id,
                    f.start_at,
                    f.name AS feature_name,
                    COUNT(s.scenario_run_id) AS scenarios,
                    SUM(CASE WHEN s.status = 'passed' THEN 1 ELSE 0 END) AS passed,
                    SUM(CASE WHEN s.status = 'failed' THEN 1 ELSE 0 END) AS failed,
                    SUM(CASE WHEN s.status = 'skipped' THEN 1 ELSE 0 END) AS skipped,
                    SUM(CASE WHEN s.status = 'errored' THEN 1 ELSE 0 END) AS errored,
                    SUM(s.duration) AS duration
                FROM cucu_run r
                JOIN worker w ON r.cucu_run_id = w.cucu_run_id
                JOIN feature f ON w.worker_run_id = f.worker_run_id
                JOIN scenario s ON f.feature_run_id = s.feature_run_id
                GROUP BY f.feature_run_id
            )
            SELECT
                *,
                CASE
                    WHEN failed  > 0 THEN 'failed'
                    WHEN errored > 0 THEN 'errored'
                    WHEN passed  > 0 THEN 'passed'
                    WHEN skipped > 0 THEN 'skipped'
                END AS status
            FROM feature_first_level
            ORDER BY start_at ASC
        """)
    db.execute_sql("""
            CREATE VIEW IF NOT EXISTS flat AS
            SELECT
                w.cucu_run_id,
                s.start_at,
                s.duration,
                f.name AS feature_name,
                s.name AS scenario_name,
                CASE
                    WHEN f.tags = '[]' AND s.tags = '[]' THEN JSON('[]')
                    WHEN f.tags = '[]' THEN s.tags
                    WHEN s.tags = '[]' THEN f.tags
                    ELSE JSON(REPLACE(f.tags, ']', '') || ',' || REPLACE(s.tags, '[', ''))
                END as tags,
                s.log_files
            FROM scenario s
            JOIN feature f ON s.feature_run_id = f.feature_run_id
            JOIN worker w ON f.worker_run_id = w.worker_run_id
        """)
    db.execute_sql("""
            CREATE VIEW IF NOT EXISTS flat_scenario AS
            SELECT
                s.scenario_run_id,
                f.name AS feature_name,
                s.name AS scenario_name,
                CASE
                    WHEN f.tags = '[]' AND s.tags = '[]' THEN JSON('[]')
                    WHEN f.tags = '[]' THEN s.tags
                    WHEN s.tags = '[]' THEN f.tags
                    ELSE JSON(REPLACE(f.tags, ']', '') || ',' || REPLACE(s.tags, '[', ''))
                END as tags,
                f.filename || ':' || s.line_number AS feature_file_line,
                s.status,
                (
                    SELECT json_group_array(json_object(
                        'status', st.status,
                        'duration', st.duration,
                        'name', st.name
                    ))
                    FROM step st
                    WHERE st.scenario_run_id = s.scenario_run_id
                    ORDER BY st.seq
                ) AS steps,
                (
                    SELECT st.debug_output
                    FROM step st
                    WHERE st.scenario_run_id = s.scenario_run_id
                    ORDER BY st.seq DESC
                    LIMIT 1
                ) AS last_step_debug_log
            FROM scenario s
            JOIN feature f ON s.feature_run_id = f.feature_run_id
        """)


def get_first_cucu_run_filepath():
    run_record = cucu_run.select().first()
    return run_record.filepath


def consolidate_database_files(results_dir, combine=False):
    # This function would need a more advanced approach with peewee, so for now, keep using sqlite3 for consolidation
    results_path = Path(results_dir)
    target_db_path = results_path / "run.db"
    if not target_db_path.exists():
        cucu_logger.info(
            f"Creating new consolidated database at {target_db_path}"
        )
        create_database_file(target_db_path)
    else:
        cucu_logger.debug(f"Found existing database at {target_db_path}")

    if not combine:
        db_files = [
            db for db in results_path.glob("**/run*.db") if db.name != "run.db"
        ]
    else:
        # include all run.db files in all subdirectories
        db_files = [
            db for db in results_path.rglob("run*.db") if db != Path("run.db")
        ]

    if not db_files:
        cucu_logger.debug("No database files found to consolidate.")
        return
    else:
        cucu_logger.debug(
            f"Found {len(db_files)} database files to consolidate."
        )

    tables_to_copy = ["cucu_run", "worker", "feature", "scenario", "step"]
    with sqlite3.connect(target_db_path) as target_conn:
        target_cursor = target_conn.cursor()
        for db_file in db_files:
            with sqlite3.connect(db_file) as source_conn:
                source_cursor = source_conn.cursor()
                for table_name in tables_to_copy:
                    source_cursor.execute(f"SELECT * FROM {table_name}")
                    rows = source_cursor.fetchall()
                    source_cursor.execute(f"PRAGMA table_info({table_name})")
                    columns = [col[1] for col in source_cursor.fetchall()]

                    # prep cucu_run for combining multiple runs
                    if table_name == "cucu_run":
                        db_path_index = columns.index("db_path")
                        rows = [
                            tuple(
                                item if idx != db_path_index else str(db_file)
                                for idx, item in enumerate(row)
                            )
                            for row in rows
                        ]

                    placeholders = ",".join(["?" for _ in columns])
                    target_cursor.executemany(
                        f"INSERT OR REPLACE INTO {table_name} VALUES ({placeholders})",
                        rows,
                    )
                    target_conn.commit()

            if not combine and db_file.name != "run.db":
                # remove the worker db files
                db_file.unlink()


def init_html_report_db(db_path):
    db.init(db_path)
    db.connect(reuse_if_open=True)


def close_html_report_db():
    db.close()



# -*- coding: utf-8 -*-
import os
import traceback
from datetime import datetime
from xml.sax.saxutils import escape

import bs4
from behave.formatter.base import Formatter
from behave.model_core import Status
from bs4.formatter import XMLFormatter
from tenacity import RetryError

from cucu.ansi_parser import remove_ansi
from cucu.config import CONFIG
from cucu.utils import ellipsize_filename


class CucuJUnitFormatter(Formatter):
    name = "cucu-junit"
    description = "JUnit Formater"

    def __init__(self, stream_opener, config):
        super(CucuJUnitFormatter, self).__init__(stream_opener, config)
        self.xml_root = None
        self.current_feature = None
        self.current_scenario = None
        self.curent_step = None
        self.current_scenario_traceback = None
        self.current_scenario_duration = 0.0
        self.current_scenario_results = {}
        self.feature_timestamp = None
        self.steps = []

    # -- FORMATTER API:
    def uri(self, uri):
        pass

    def feature(self, feature):
        self.current_feature = feature
        date_now = datetime.now()
        self.feature_results = {
            "name": escape(feature.name),
            "foldername": escape(ellipsize_filename(feature.name)),
            "tests": 0,
            "errors": 0,
            "failures": 0,
            "skipped": 0,
            "timestamp": date_now.strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
            "scenarios": {},
        }
        if feature.tags:
            self.feature_results["tags"] = ", ".join(feature.tags)

    def background(self, background):
        # -- ADD BACKGROUND STEPS: Support *.feature file regeneration.
        for step_ in background.steps:
            self.step(step_)

    def update_scenario(self):
        if self.current_scenario is not None:
            self.current_scenario_results["time"] = str(
                round(self.current_scenario_duration, 3)
            )
            hook_failed = self.current_scenario.hook_failed
            if hook_failed:
                status = "errored"
            else:
                status = self.current_scenario.compute_status().name

            self.current_scenario_results["status"] = status
            failures = []

            if status == "failed" and self.current_scenario_traceback:
                failure_handlers = CONFIG["__CUCU_CUSTOM_FAILURE_HANDLERS"]

                for failure_handler in failure_handlers:
                    failures.append(
                        failure_handler(
                            self.current_feature, self.current_scenario
                        )
                    )

                failures += [
                    f"{self.current_step.keyword} {self.current_step.name} (after {round(self.current_step.duration, 3)}s)"
                ]

                if error := self.current_step.exception:
                    if isinstance(error, RetryError):
                        error = error.last_attempt.exception()

                    if len(error.args) > 0 and isinstance(error.args[0], str):
                        error_class_name = error.__class__.__name__
                        error_lines = error.args[0].splitlines()
                        error_lines[0] = (
                            f"{error_class_name}: {error_lines[0]}"
                        )
                    else:
                        error_lines = [repr(error)]
                    failures += error_lines

                if CONFIG["CUCU_JUNIT_WITH_STACKTRACE"] == "true":
                    failures += traceback.format_tb(
                        self.current_scenario_traceback
                    )

                self.current_scenario_results["failure"] = failures

            if status == "skipped":
                self.current_scenario_results["skipped"] = True

    def scenario(self, scenario):
        self.update_scenario()

        self.current_scenario = scenario
        self.current_scenario_traceback = None
        self.current_scenario_duration = 0.0
        self.current_scenario_results = {
            "status": "pending",
            "time": "n/a",
            "failure": None,
            "skipped": None,
        }
        self.current_scenario_results["foldername"] = escape(
            ellipsize_filename(scenario.name)
        )
        if scenario.tags:
            self.current_scenario_results["tags"] = ", ".join(scenario.tags)

        scenario_name = escape(scenario.name)
        self.feature_results["scenarios"][scenario_name] = (
            self.current_scenario_results
        )

        # we write out every new scenario into the JUnit results output
        # which allows us to have a valid JUnit XML results file per feature
        # file currently running even if the process crashes or is killed so we
        # can still generate valid reports from every run.
        self.write_results(self.feature_results)

    def match(self, step):
        pass

    def step(self, step):
        self.steps.append(step)

    def insert_step(self, step, index=-1):
        if index == -1:
            self.steps.append(step)
        else:
            self.steps.insert(index, step)

    def result(self, step):
        self.current_step = step
        if step.status == Status.failed:
            self.current_scenario_traceback = step.exc_traceback

        self.current_scenario_duration += step.duration

    def eof(self):
        """
        end of file for the feature and this is when we write the updated
        results to the file
        """
        self.update_scenario()
        self.write_results(self.feature_results)

    def close(self):
        pass

    def write_results(self, results):
        """
        given a feature results dictionary that looks like so:
            {
                "name": "name of the feature",
                "foldername": "",
                "tests": 0,
                "errors": 0,
                "failures": 0,
                "skipped": 0,
                "timestamp": "",
                "scenarios": {
                    "scenario name": {
                        "foldername": "",
                        "tags": "DOM-3435, testrail(3366,45891)",
                        "status": "passed/failed/skipped",
                        "time": "0.0000":
                        "stdout": "",
                        "stderr": "",
                    },
                    ...
                }
            }
        """

        # custom attribute ordering so attributes are printed in a consistent
        # and desired order
        class SortAttributes(XMLFormatter):
            def attributes(self, tag):
                ordered = [
                    "classname",
                    "name",
                    "foldername",
                    "tests",
                    "errors",
                    "failures",
                    "skipped",
                    "status",
                    "timestamp",
                    "time",
                    "tags",
                ]

                return [
                    (attr, tag[attr]) for attr in ordered if attr in tag.attrs
                ]

        soup = bs4.BeautifulSoup()
        testsuite = bs4.Tag(name="testsuite")
        testsuite["name"] = results["name"]
        testsuite["foldername"] = results["foldername"]
        testsuite["timestamp"] = results["timestamp"]

        junit_dir = CONFIG["CUCU_JUNIT_DIR"]
        os.makedirs(junit_dir, exist_ok=True)

        feature_name = results["name"]
        output_filepath = os.path.join(junit_dir, f"{feature_name}.xml")

        scenarios = results["scenarios"]

        if CONFIG["CUCU_SHOW_SKIPS"] != "true":
            filtered_scenarios = {}

            for name, scenario in scenarios.items():
                if scenario["status"] != "skipped":
                    filtered_scenarios[name] = scenario

            scenarios = filtered_scenarios

            if len(scenarios) == 0:
                # we had a suite of just skipped results
                return

        # calculate with the latest data
        # workaround for beatufulsoup4 removing the attribute if it is set to 0
        testsuite["tests"] = str(len(scenarios))
        testsuite["failures"] = str(
            len(
                [x for x in scenarios.values() if x["status"] == Status.failed]
            )
        )
        testsuite["skipped"] = str(
            len(
                [
                    x
                    for x in scenarios.values()
                    if x["status"] == Status.skipped
                ]
            )
        )
        testsuite["errors"] = str(
            len(
                [
                    x
                    for x in scenarios.values()
                    if x["status"]
                    not in (Status.failed, Status.skipped, Status.passed)
                ]
            )
        )

        if "tags" in results:
            testsuite["tags"] = results["tags"]
        soup.append(testsuite)

        for scenario_name in scenarios:
            scenario = scenarios[scenario_name]
            testcase = bs4.Tag(name="testcase")
            testcase["classname"] = results["name"]
            testcase["name"] = scenario_name
            testcase["foldername"] = scenario["foldername"]
            if "tags" in scenario:
                testcase["tags"] = scenario["tags"]
            testcase["status"] = scenario["status"]
            testcase["time"] = scenario["time"]

            if scenario["failure"] is not None:
                failure_message = "\n".join(scenario["failure"])
                failure = bs4.Tag(name="failure")
                cleaned_failure_message = remove_ansi(failure_message)
                failure.append(bs4.CData(cleaned_failure_message))
                testcase.append(failure)

            if scenario["skipped"] is not None:
                testcase.append(bs4.Tag(name="skipped"))

            testsuite.append(testcase)

        feature_name = results["name"]
        output_filepath = os.path.join(junit_dir, f"{feature_name}.xml")
        with open(output_filepath, "w", encoding="utf-8") as output:
            output.write(soup.prettify(formatter=SortAttributes()))



# import all of the steps from cucu
from cucu.steps import *  # noqa: F403, F401

# import individual sub-modules here (i.e. module names of your custom step py files)
# Example: For file features/steps/ui/login.py
# import steps.ui.login_steps
import steps.my_steps



import sys

from cucu import run_steps, step


@step("I skip this scenario if not on mac")
def skip_scenario_if_on_mac(ctx):
    if "darwin" not in sys.platform:
        ctx.scenario.skip(reason="skipping scenario since we're not on a mac")


@step("I skip this scenario if not on linux")
def skip_scenario_if_on_linuxc(ctx):
    if "linux" not in sys.platform:
        ctx.scenario.skip(
            reason="skipping scenario since we're not on a linux"
        )


@step("I run the following steps if on mac")
def run_steps_if_on_mac(ctx):
    if "darwin" in sys.platform:
        run_steps(ctx, ctx.text)


@step("I run the following steps if on linux")
def run_steps_if_on_linux(ctx):
    if "linux" in sys.platform:
        run_steps(ctx, ctx.text)



import base64
import os
import re

from selenium.webdriver.common.keys import Keys

from cucu import config, logger, retry, run_steps, step
from cucu.browser.selenium import Selenium


def open_browser(ctx):
    browser_name = config.CONFIG["CUCU_BROWSER"]
    headless = config.CONFIG["CUCU_BROWSER_HEADLESS"]
    selenium_remote_url = config.CONFIG["CUCU_SELENIUM_REMOTE_URL"]

    browser = Selenium()
    logger.debug(f"opening browser {browser_name}")
    browser.open(
        browser_name,
        headless=headless,
        selenium_remote_url=selenium_remote_url,
    )

    return browser


@step('I open a browser at the url "{url}"')
def open_a_browser(ctx, url):
    """
    open a browser at the url provided

    example:
        Given I open a browser at the url "https://www.google.com"
    """
    if ctx.browser is None:
        ctx.browser = retry(open_browser)(ctx)
        ctx.browsers.append(ctx.browser)
    else:
        logger.debug("browser already open so using existing instance")

    logger.debug(f"navigating to url: {url}")
    ctx.browser.navigate(url)


@step('I open a new browser at the url "{url}"')
def open_a_new_browser(ctx, url):
    ctx.browser = retry(open_browser)(ctx)
    ctx.browsers.append(ctx.browser)
    logger.debug(f"navigating to url: {url}")
    ctx.browser.navigate(url)


@step("I execute in the current browser the following javascript")
def execute_javascript(ctx):
    ctx.check_browser_initialized()
    ctx.browser.execute(ctx.text)


@step(
    'I execute in the current browser the following javascript and save the result to the variable "{variable}"'
)
def execute_javascript_and_save(ctx, variable):
    ctx.check_browser_initialized()
    result = ctx.browser.execute(ctx.text)
    config.CONFIG[variable] = result


def assert_url_is(ctx, value):
    ctx.check_browser_initialized()
    url = ctx.browser.get_current_url()
    if value != url:
        raise RuntimeError(f"current url is {url}, not {value}")


@step('I should see the current url is "{value}"')
def should_see_the_current_url_is(ctx, value):
    assert_url_is(ctx, value)


@step('I wait to see the current url is "{value}"')
def wait_to_see_the_current_url_is(ctx, value):
    retry(assert_url_is)(ctx, value)


def assert_url_matches(ctx, regex):
    ctx.check_browser_initialized()
    url = ctx.browser.get_current_url()
    if not re.search(regex, url):
        raise RuntimeError(f"current url {url} does not match {regex}")


@step('I should see the current url matches "{regex}"')
def should_see_the_current_url_matches(ctx, regex):
    assert_url_matches(ctx, regex)


@step('I wait to see the current url matches "{regex}"')
def wait_to_see_the_current_url_matches(ctx, regex):
    retry(assert_url_matches)(ctx, regex)


@step('I save the current url to the variable "{variable}"')
def save_current_url_to_variable(ctx, variable):
    ctx.check_browser_initialized()
    config.CONFIG[variable] = ctx.browser.get_current_url()


@step("I refresh the browser")
def refresh_browser(ctx):
    ctx.check_browser_initialized()
    ctx.browser.refresh()


@step("I go back on the browser")
def go_back_on_browser(ctx):
    ctx.check_browser_initialized()
    ctx.browser.back()


@step('I save the contents of the clipboard to the variable "{variable}"')
def save_clipboard_value_to_variable(ctx, variable):
    ctx.check_browser_initialized()

    # use default frame when adding elements to the document to avoid this errro on access:
    # "selenium.common.exceptions.ElementNotInteractableException: Message: element not interactable"
    ctx.browser.switch_to_default_frame()

    # create the hidden textarea so we can paste clipboard contents in
    textarea = ctx.browser.execute(
        """
    var textarea = document.getElementById('cucu-copy-n-paste')
    if (!textarea) {
        textarea = document.createElement('textarea');
        textarea.setAttribute('id', 'cucu-copy-n-paste');
        textarea.style.display = 'hidden';
        textarea.style.height = '0px';
        textarea.style.width = '0px';
        document.body.insertBefore(textarea, document.body.firstChild);
    }
    return textarea;
    """
    )

    # send ctrl+v or cmd+v to that element
    if "mac" in ctx.browser.execute("return navigator.platform").lower():
        textarea.send_keys(Keys.COMMAND, "v")
    else:
        textarea.send_keys(Keys.CONTROL, "v")

    clipboard_contents = textarea.get_attribute("value")
    config.CONFIG[variable] = clipboard_contents


@step('I should see the browser title is "{title}"')
def should_see_browser_title(ctx, title):
    ctx.check_browser_initialized()
    current_title = ctx.browser.title()

    if current_title != title:
        raise RuntimeError(f'unexpected browser title, got "{current_title}"')


@step("I close the current browser")
def close_browser(ctx):
    ctx.check_browser_initialized()
    browser_index = ctx.browsers.index(ctx.browser)

    if browser_index > 0:
        ctx.browser = ctx.browsers[browser_index - 1]
    else:
        ctx.browser = None

    ctx.browsers[browser_index].quit()
    del ctx.browsers[browser_index]


@step('I navigate to the url "{url}"')
def navigate_to_the_url(ctx, url):
    ctx.check_browser_initialized()
    logger.debug(f"navigating to url: {url}")
    ctx.browser.navigate(url)


@step('I save the current browser url to the variable "{variable}"')
def save_current_browser_url_to_variable(ctx, variable):
    ctx.check_browser_initialized()
    current_url = ctx.browser.get_current_url()
    config.CONFIG[variable] = current_url
    logger.debug(
        f"saved current browser url {current_url} to variable {variable}"
    )


@step("I switch to the previous browser")
def switch_to_previous_browser(ctx):
    browser_index = ctx.browsers.index(ctx.browser)

    if browser_index > 0:
        ctx.browser = ctx.browsers[browser_index - 1]
    else:
        raise RuntimeError("no previous browser window available")


@step("I switch to the next browser")
def switch_to_next_browser(ctx):
    browser_index = ctx.browsers.index(ctx.browser)

    if browser_index < len(ctx.browsers) - 1:
        ctx.browser = ctx.browsers[browser_index + 1]
    else:
        raise RuntimeError("no next browser window available")


@step("I close the current browser tab")
def close_browser_tab(ctx):
    ctx.check_browser_initialized()
    ctx.browser.close_window()


def switch_to_next_tab(ctx):
    ctx.check_browser_initialized()
    ctx.browser.switch_to_next_tab()


@step("I switch to the next browser tab")
def switch_to_next_browser_tab(ctx):
    switch_to_next_tab(ctx)


@step("I wait to switch to the next browser tab")
def wait_to_switch_to_next_browser_tab(ctx):
    retry(switch_to_next_tab)(ctx)


def switch_to_previous_tab(ctx):
    ctx.check_browser_initialized()
    ctx.browser.switch_to_previous_tab()


@step("I switch to the previous browser tab")
def switch_to_previous_browser_tab(ctx):
    switch_to_previous_tab(ctx)


@step("I wait to switch to the previous browser tab")
def wait_to_switch_to_previous_browser_tab(ctx):
    retry(switch_to_previous_tab)(ctx)


def switch_to_nth_tab(ctx, tab_number):
    ctx.check_browser_initialized()
    ctx.browser.switch_to_nth_tab(tab_number)


@step('I switch to the "{nth:nth}" browser tab')
def switch_to_nth_browser_tab(ctx, nth):
    switch_to_nth_tab(ctx, nth)


@step('I wait to switch to the "{nth:nth}" browser tab')
def wait_to_switch_to_nth_browser_tab(ctx, nth):
    retry(switch_to_nth_tab)(ctx, nth)


def switch_to_tab_matching_regex(ctx, regex):
    ctx.check_browser_initialized()
    ctx.browser.switch_to_tab_that_matches_regex(regex)


@step('I switch to the browser tab that matches "{regex}"')
def switch_to_browser_tab_matching_regex(ctx, regex):
    switch_to_tab_matching_regex(ctx, regex)


@step('I wait to switch to the browser tab that matches "{regex}"')
def wait_to_switch_to_browser_tab_matching_regex(ctx, regex):
    retry(switch_to_tab_matching_regex)(ctx, regex)


@step('I save the browser tabs info to the variable "{variable}"')
def save_browser_tabs_info_to_variable(ctx, variable):
    ctx.check_browser_initialized()
    tabs_info = ctx.browser.get_all_tabs_info()
    lst_tab_info = [
        f"tab({index + 1}): {tab['title']}, url: {tab['url']}"
        for index, tab in enumerate(tabs_info)
    ]
    config.CONFIG[variable] = lst_tab_info


@step("I list the current browser tab info")
def list_current_browser_tab_info(ctx):
    ctx.check_browser_initialized()
    tab_info = ctx.browser.get_tab_info()
    current_tab = tab_info["index"] + 1
    title = tab_info["title"]
    url = tab_info["url"]
    print(f"\ntab({current_tab}): {title}\nurl: {url}\n")


@step("I list all browser tabs info")
def list_browser_tabs_info(ctx):
    ctx.check_browser_initialized()
    all_tabs_info = ctx.browser.get_all_tabs_info()
    for index, tab in enumerate(all_tabs_info):
        tab_index = index + 1
        title = tab["title"]
        url = tab["url"]
        print(f"\ntab({tab_index}): {title}\nurl: {url}\n")


def save_downloaded_file(ctx, filename):
    ctx.check_browser_initialized()

    # use default frame when adding elements to the document to avoid this errro on access:
    # "selenium.common.exceptions.ElementNotInteractableException: Message: element not interactable"
    ctx.browser.switch_to_default_frame()

    elem = ctx.browser.execute(
        """
        var input = window.document.createElement('INPUT');
        input.setAttribute('type', 'file');
        input.hidden = true;
        input.onchange = function (e) { e.stopPropagation() };
        return window.document.documentElement.appendChild(input);
        """
    )
    cucu_downloads_dir = config.CONFIG["CUCU_BROWSER_DOWNLOADS_DIR"]
    elem.send_keys(f"{cucu_downloads_dir}/{filename}")
    ctx.browser.execute(
        """
        var input = arguments[0];
        window.__cucu_downloaded_file = null;
        var reader = new FileReader();
        reader.onload = function (ev) {
            window.__cucu_downloaded_file = reader.result;
        };
        reader.onerror = function (ex) {
            window.__cucu_downloaded_file = ex.message;
        };
        reader.readAsDataURL(input.files[0]);
        input.remove();
        """,
        elem,
    )

    def wait_for_file():
        if (
            ctx.browser.execute("return window.__cucu_downloaded_file;")
            is None
        ):
            raise RuntimeError(f"waiting on file {filename}")

    retry(wait_for_file)()

    result = ctx.browser.execute("return window.__cucu_downloaded_file;")
    if not result.startswith("data:"):
        raise Exception("Failed to get file content: %s" % result)

    filedata = base64.b64decode(result[result.find("base64,") + 7 :])
    scenario_downloads_dir = config.CONFIG["SCENARIO_DOWNLOADS_DIR"]
    download_filepath = os.path.join(scenario_downloads_dir, filename)
    open(download_filepath, "wb").write(filedata)


@step('I wait to see the downloaded file "{filename}"')
def wait_to_see_downloaded_file(ctx, filename):
    """
    wait to see the expected downloaded filename appears in the current
    browsers download directory and internally we then copy the contents of
    that file to the SCENARIO_DOWNLOADS_DIR so the test can continue to
    use the file as it deems necessary.
    """
    retry(save_downloaded_file)(ctx, filename)


@step(
    'I wait up to "{seconds}" seconds to see the downloaded file "{filename}"'
)
def wait_up_to_seconds_to_see_downloaded_file(ctx, seconds, filename):
    seconds = float(seconds)
    retry(save_downloaded_file, wait_up_to_s=seconds)(ctx, filename)


@step('I download an mht archive of the current page to "{file_path}"')
def download_mht_archive(ctx, file_path):
    ctx.browser.download_mht(file_path)


@step('I run the following steps if the current browser is "{name}"')
def run_if_browser(ctx, name):
    if config.CONFIG["CUCU_BROWSER"].lower() == name.lower():
        run_steps(ctx, ctx.text)


@step('I do not run the following steps if the current browser is "{name}"')
def run_if_not_browser(ctx, name):
    if config.CONFIG["CUCU_BROWSER"].lower() != name.lower():
        run_steps(ctx, ctx.text)


@step('I skip this scenario if the current browser is "{name}"')
def skip_if_browser(ctx, name):
    if config.CONFIG["CUCU_BROWSER"].lower() == name.lower():
        ctx.scenario.skip(reason=f"skipping scenario since we're on {name}")


@step('I skip this scenario if the current browser is not "{name}"')
def skip_if_not_browser(ctx, name):
    if config.CONFIG["CUCU_BROWSER"].lower() != name.lower():
        ctx.scenario.skip(
            reason=f"skipping scenario since we're not on {name}"
        )


@step('I save the browser cookie "{cookie_name}" to the variable "{variable}"')
def save_browser_cookie(ctx, cookie_name, variable):
    ctx.check_browser_initialized()
    config.CONFIG[variable] = ctx.browser.driver.get_cookie(cookie_name)[
        "value"
    ]


@step('I set the browser cookie "{name}" a value of "{value}"')
def add_browser_cookie(ctx, name, value):
    ctx.check_browser_initialized()
    ctx.browser.driver.add_cookie({"name": name, "value": value})


# step to get the driver window size
@step('I get the browser window size and save it to the variable "{variable}"')
def get_browser_window_size(ctx, variable):
    ctx.check_browser_initialized()
    size = ctx.browser.driver.get_window_size()
    config.CONFIG[variable] = f"{size['width']}x{size['height']}"



import logging
import os
import re

import chromedriver_autoinstaller
import geckodriver_autoinstaller
import urllib3
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.remote.command import Command

from cucu import config, edgedriver_autoinstaller, logger
from cucu.browser.core import Browser
from cucu.browser.frames import search_in_all_frames


class DisableLogger:
    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


def init():
    """
    initialize any selenium specific needs
    """
    if config.CONFIG["CUCU_BROWSER"] == "chrome":
        try:
            with DisableLogger():
                # auto install chromedriver if not present
                chromedriver_autoinstaller.install()
        except:  # noqa: E722
            logging.debug("unable to auto install chromedriver")

    if config.CONFIG["CUCU_BROWSER"] == "firefox":
        # https://github.com/mozilla/geckodriver/issues/330
        logger.warning("browser console logs not available on firefox")
        geckodriver_autoinstaller.install()

    if config.CONFIG["CUCU_BROWSER"] == "edge":
        edgedriver_autoinstaller.install()


class Selenium(Browser):
    def __init__(self):
        self.driver = None

    def open(
        self, browser, headless=False, selenium_remote_url=None, detach=False
    ):
        if selenium_remote_url is None:
            init()

        timeout = float(config.CONFIG["CUCU_SELENIUM_DEFAULT_TIMEOUT_S"])
        os.environ["GLOBAL_DEFAULT_TIMEOUT"] = str(timeout)

        height = config.CONFIG["CUCU_BROWSER_WINDOW_HEIGHT"]
        width = config.CONFIG["CUCU_BROWSER_WINDOW_WIDTH"]
        cucu_downloads_dir = config.CONFIG["CUCU_BROWSER_DOWNLOADS_DIR"]
        ignore_ssl_certificate_errors = config.CONFIG[
            "CUCU_IGNORE_SSL_CERTIFICATE_ERRORS"
        ]

        if browser.startswith("chrome"):
            options = ChromeOptions()
            options.add_experimental_option("detach", detach)

            prefs = {"download.default_directory": cucu_downloads_dir}
            options.add_experimental_option("prefs", prefs)

            options.add_argument(f"--window-size={width},{height}")
            options.add_argument("--disable-dev-shm-usage")

            # Disable the automation message to get back some vertical space
            options.add_argument("--disable-infobars")  # old way
            options.add_experimental_option(
                "excludeSwitches", ["enable-automation"]
            )  # new way

            if headless:
                options.add_argument("--headless")

            if ignore_ssl_certificate_errors:
                options.add_argument("ignore-certificate-errors")
            options.set_capability("goog:loggingPrefs", {"browser": "ALL"})

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Chrome init")
                self.driver = webdriver.Chrome(
                    options=options,
                )

        elif browser.startswith("firefox"):
            options = FirefoxOptions()

            options.set_preference("browser.download.folderList", 2)
            options.set_preference(
                "browser.download.manager.showWhenStarting", False
            )
            options.set_preference("browser.download.dir", cucu_downloads_dir)
            options.set_preference(
                "browser.helperApps.neverAsk.saveToDisk",
                "images/jpeg, application/pdf, application/octet-stream, text/plain",
            )

            if ignore_ssl_certificate_errors:
                options.accept_insecure_certs = True

            options.add_argument(f"--width={width}")
            options.add_argument(f"--height={height}")

            # TODO: re-enable once Firefox supports this
            # options.set_capability("loggingPrefs", {"browser": "ALL"})

            if headless:
                options.add_argument("--headless")

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Firefox init")
                self.driver = webdriver.Firefox(
                    options=options,
                )

        elif browser.startswith("edge"):
            options = EdgeOptions()

            options.add_experimental_option(
                "prefs", {"download.default_directory": cucu_downloads_dir}
            )

            if headless:
                options.use_chromium = True
                options.add_argument("--headless")

            if ignore_ssl_certificate_errors:
                options.set_capability("acceptSslCerts", True)

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Edge init")
                edgedriver_filepath = (
                    edgedriver_autoinstaller.utils.download_edgedriver()
                )
                self.driver = webdriver.Edge(
                    executable_path=edgedriver_filepath,
                    options=options,
                )

        else:
            raise Exception(f"unknown browser {browser}")

        self.driver.set_window_size(width, height)
        session_id = self.get_session_id()
        logger.debug(f"cucu started Selenium session with ID: {session_id}")

    def get_session_id(self):
        if self.driver:
            return getattr(self.driver, "session_id", None)
        return None

    def get_log(self):
        if config.CONFIG["CUCU_BROWSER"] == "firefox":
            return []

        try:
            # local selenium
            return self.driver.get_log("browser")
        except AttributeError:
            # remote selemium
            # workaround from https://github.com/SeleniumHQ/selenium/issues/15772#issuecomment-3357920550
            return self.driver.execute(Command.GET_LOG, {"type": "browser"})[
                "value"
            ]

    def get_current_url(self):
        return self.driver.current_url

    def navigate(self, url):
        self.driver.get(url)
        self.wait_for_page_to_load()

    def switch_to_next_tab(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        if window_handle_index == len(window_handles) - 1:
            raise RuntimeError("no next browser tab available")
        self.driver.switch_to.window(window_handles[window_handle_index + 1])

    def switch_to_previous_tab(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        if window_handle_index == 0:
            raise RuntimeError("no previous browser tab available")
        self.driver.switch_to.window(window_handles[window_handle_index - 1])

    def switch_to_nth_tab(self, tab_number):
        print(f"tab_number, {tab_number}")
        window_handles = self.driver.window_handles
        total_tabs = len(window_handles)
        if tab_number > total_tabs:
            raise RuntimeError(f"no {tab_number} browser tab available")
        self.driver.switch_to.window(window_handles[tab_number])

    def switch_to_tab_that_matches_regex(self, title_pattern):
        window_handles = self.driver.window_handles

        for handle in window_handles:
            self.driver.switch_to.window(handle)
            if re.search(title_pattern, self.driver.title):
                return

        raise Exception(f"No tab title matches pattern: {title_pattern}")

    def get_tab_info(self):
        window_handles = self.driver.window_handles
        current_window = self.driver.current_window_handle
        window_handle_index = window_handles.index(current_window)

        return {
            "tab_count": len(window_handles),
            "index": window_handle_index,
            "title": self.driver.title,
            "url": self.driver.current_url,
        }

    def get_all_tabs_info(self):
        window_handles = self.driver.window_handles
        current_window = self.driver.current_window_handle
        tabs_info = []

        for handle in window_handles:
            self.driver.switch_to.window(handle)
            tabs_info.append(
                {
                    "title": self.driver.title,
                    "url": self.driver.current_url,
                }
            )

        # Switch back to the original window
        self.driver.switch_to.window(current_window)

        return tabs_info

    def back(self):
        self.driver.back()
        self.wait_for_page_to_load()

    def refresh(self):
        self.driver.refresh()
        self.wait_for_page_to_load()

    def title(self):
        return self.driver.title

    def css_find_elements(self, selector):
        def find_elements_in_frame():
            elements = self.driver.find_elements(By.CSS_SELECTOR, selector)

            def visible(element):
                return element.is_displayed()

            return list(filter(visible, elements))

        return search_in_all_frames(self, find_elements_in_frame)

    def execute(self, javascript, *args):
        return self.driver.execute_script(javascript, *args)

    def click(self, element):
        # if Firefox just do normal click
        if config.CONFIG["CUCU_BROWSER"] == "firefox":
            element.click()
        else:
            actions = ActionChains(self.driver)
            actions.move_to_element(element).click().perform()

        # let cucu's own wait for page to load checks run
        self.wait_for_page_to_load()

    def switch_to_default_frame(self):
        self.driver.switch_to.default_content()

    def switch_to_frame(self, frame):
        self.driver.switch_to.frame(frame)

    def screenshot(self, filepath):
        self.driver.get_screenshot_as_file(filepath)

    def close_window(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        self.driver.close()
        self.driver.switch_to.window(window_handles[window_handle_index - 1])

    def download_mht(self, target_filepath):
        if self.driver is None:
            logger.warning(
                "No active browser; will not attempt to download .mht file."
            )
            return

        browser_name = self.driver.name.lower()
        if "chrome" not in browser_name:
            logger.warning(
                "The web driver is not using Chrome as a web browser"
                f", but {browser_name}. This browser does not support"
                "dowloading .mht files; will not attempt to download one."
            )
            return

        mht_data = None
        if self.driver._is_remote:
            cdp_url = f"{self.driver.command_executor._client_config.remote_server_addr}/session/{self.driver.session_id}/chromium/send_command_and_get_result"
            cdp_request_body = '{"cmd": "Page.captureSnapshot", "params": {}}'
            cdp_response = self.driver.command_executor._request(
                "POST", cdp_url, cdp_request_body
            )
            # Flaky: Adding try catch block to handle the situation where we do not get a dictionary object.
            # Could not reproduce this in local. So far we have seen this issue only when running on a remote web driver.
            try:
                mht_data = cdp_response.get("value").get("data")
            except Exception as e:
                logger.error(f'object "{cdp_response.get("value")}" : {e}')
        else:
            mht_response = self.driver.execute_cdp_cmd(
                "Page.captureSnapshot", {}
            )
            mht_data = mht_response.get("data")

        if mht_data is None:
            logger.warning(
                "Something unexpected has happened: fetched MHT data, but that data was empty. Not writing MHT file."
            )
            return

        with open(target_filepath, "w") as mht_file:
            mht_file.write(mht_data)

    def quit(self):
        self.driver.quit()



import os

import humanize

from cucu import fuzzy, logger, retry, step
from cucu.utils import take_saw_element_screenshot


def find_file_input(ctx, name, index=0):
    """

        * <input type="file">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired button on screen
      index(str):  the index of the button if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser, name, ['input[type="file"]'], index=index
    )

    prefix = "" if index == 0 else f"{humanize.ordinal(index)} "

    take_saw_element_screenshot(ctx, "file input", name, index, element)

    if element is None:
        raise RuntimeError(f'unable to find the {prefix}file input "{name}"')

    return element


@step('I upload the file "{filepath}" to the file input "{name}"')
def upload_file_to_input(ctx, filepath, name):
    _input = find_file_input(ctx, name)
    _input.send_keys(os.path.abspath(filepath))


JS_DROP_FILE = """
    var args = arguments,
    element = args[0],
    offsetX = args[1],
    offsetY = args[2],
    doc = element.ownerDocument || document;

    var box = element.getBoundingClientRect(),
        clientX = box.left + (offsetX || box.width / 2),
        clientY = box.top + (offsetY || box.height / 2),
        target = doc.elementFromPoint(clientX, clientY);
    if (!(target && element.contains(target))) {
        element.scrollIntoView({
            behavior: "instant",
            block: "center",
            inline: "center",
        });
        var box = element.getBoundingClientRect(),
            clientX = box.left + (offsetX || box.width / 2),
            clientY = box.top + (offsetY || box.height / 2),
            target = doc.elementFromPoint(clientX, clientY);

        if (!(target && element.contains(target))) {
            // Scrolling didn't do the trick, so give up
            var ex = new Error("Element not interactable");
            ex.code = 15;
            throw ex;
        }
    }

    var input = doc.createElement("INPUT");
    input.setAttribute("type", "file");
    input.setAttribute("multiple", "");
    input.setAttribute("style", "position:fixed;z-index:2147483647;left:0;top:0;");
    input.onchange = function (ev) {
    input.parentElement.removeChild(input);
    ev.stopPropagation();

    var dataTransfer = {
        constructor: DataTransfer,
        effectAllowed: "all",
        dropEffect: "none",
        types: ["Files"],
        files: input.files,
        setData: function setData() {},
        getData: function getData() {},
        clearData: function clearData() {},
        setDragImage: function setDragImage() {},
    };

    if (window.DataTransferItemList) {
        dataTransfer.items = Object.setPrototypeOf(
        Array.prototype.map.call(input.files, function (f) {
            return {
            constructor: DataTransferItem,
            kind: "file",
            type: f.type,
            getAsFile: function getAsFile() {
                return f;
            },
            getAsString: function getAsString(callback) {
                var reader = new FileReader();
                reader.onload = function (ev) {
                callback(ev.target.result);
                };
                reader.readAsText(f);
            },
            webkitGetAsEntry: function webkitGetAsEntry() {
                return {
                constructor: window.FileSystemEntry || window.Entry,
                name: f.name,
                fullPath: "/" + f.name,
                isFile: true,
                isDirectory: false,
                file: function file(callback) {
                    callback(f);
                },
                };
            },
            };
        }),
        {
            constructor: DataTransferItemList,
            add: function add() {},
            clear: function clear() {},
            remove: function remove() {},
        }
        );
    }

    ["dragenter", "dragover", "drop"].forEach(function (type) {
        var event = doc.createEvent("DragEvent");
        event.initMouseEvent(
        type,
        true,
        true,
        doc.defaultView,
        0,
        0,
        0,
        clientX,
        clientY,
        false,
        false,
        false,
        false,
        0,
        null
        );

        Object.setPrototypeOf(event, null);
        event.dataTransfer = dataTransfer;
        Object.setPrototypeOf(event, DragEvent.prototype);

        target.dispatchEvent(event);
    });
    };

    doc.documentElement.appendChild(input);
    input.getBoundingClientRect(); /* force reflow for Firefox */
    return input;
"""


def drag_and_drop_file(ctx, name, filepath):
    drop_target = fuzzy.find(ctx.browser, name, ["*"])
    drop_target_html = drop_target.get_attribute("outerHTML")
    logger.debug(
        f'looked for drag & drop target "{name}" and found "{drop_target_html}"'
    )
    file_input = ctx.browser.execute(JS_DROP_FILE, drop_target, 0, 0)
    file_input.send_keys(os.path.abspath(filepath))


@step('I drag and drop the file "{filepath}" to "{name}"')
def should_drag_and_drop_file(ctx, filepath, name):
    drag_and_drop_file(ctx, name, filepath)


@step('I wait to drag and drop the file "{filepath}" to "{name}"')
def wait_to_drag_and_drop_file(ctx, filepath, name):
    retry(drag_and_drop_file)(ctx, name, filepath)



import socket
from functools import partial
from http.server import HTTPServer, SimpleHTTPRequestHandler
from threading import Thread

from behave import step

from cucu import logger, register_after_this_scenario_hook
from cucu.config import CONFIG


class QuietHTTPRequestHandler(SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        return


@step(
    'I start a webserver at directory "{directory}" and save the port to the variable "{variable}"'
)
def run_webserver_for_scenario(ctx, directory, variable):
    """
    start a webserver with the root at the directory provided and save the
    port that the server is listening at to the variable name provided

    examples:
        Given I start webserver at directory "/some/path" and save the port to the variable "PORT"
          And I open a browser at the url "http://{HOST_ADDRESS}:{PORT}/somefile.html"
    """
    handler = partial(QuietHTTPRequestHandler, directory=directory)
    httpd = HTTPServer(("", 0), handler)
    thread = Thread(target=httpd.serve_forever)
    thread.start()

    _, port = httpd.server_address
    CONFIG[variable] = str(port)

    with socket.create_connection(("localhost", port), timeout=5):
        logger.debug(f"Webserver is running at {port=}")

    def shutdown_webserver(_):
        httpd.shutdown()
        thread.join()

    register_after_this_scenario_hook(shutdown_webserver)



import logging
import sys
from functools import wraps

from cucu.config import CONFIG


def init_logging(logging_level):
    """
    initialize the cucu logger
    """
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(logging_level)

    logging.getLogger().addHandler(handler)
    # set the top level logger to DEBUG while the console stream handler is
    # actually set to whatever you passed in using --logging-level which is
    # INFO by default
    logging.getLogger().setLevel(logging.DEBUG)

    logging.debug("logger initialized")

    logging.getLogger("parse").setLevel(logging.WARNING)
    logging.getLogger("selenium").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)


def init_debug_logger(output_file):
    """
    initialize a debug logger handler that runs at debug level and pushes
    all of the logs to the output file provided without affecting the logging
    of the main root logger
    """
    # assume that there's only 2 loggers the first one is the console one and
    # the second one if present is a previously set debug logger for the
    # previously executed scenario
    if len(logging.getLogger().handlers) > 1:
        logging.getLogger().removeHandler(logging.getLogger().handlers[1])

    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    handler = logging.StreamHandler(output_file)
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    logging.getLogger().addHandler(handler)


@wraps(logging.log)
def log(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    msg_level = args[0]
    if logging_level <= msg_level:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.getLogger().log(*args, **kwargs)


@wraps(logging.debug)
def debug(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    if logging_level <= logging.DEBUG:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.getLogger().debug(*args, **kwargs)


@wraps(logging.info)
def info(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    if logging_level <= logging.INFO:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.info(*args, **kwargs)


@wraps(logging.warning)
def warning(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    if logging_level <= logging.WARN:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.getLogger().warning(*args, **kwargs)


@wraps(logging.error)
def error(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    if logging_level <= logging.ERROR:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.getLogger().error(*args, **kwargs)


@wraps(logging.exception)
def exception(*args, **kwargs):
    console_handler = logging.getLogger().handlers[0]
    logging_level = console_handler.level

    if logging_level <= logging.ERROR:
        CONFIG["__CUCU_WROTE_TO_OUTPUT"] = True

    logging.getLogger().exception(*args, **kwargs)



# coding: utf-8
"""
Helper functions for filename and URL generation.
"""

import logging
import os
import re
import subprocess
import sys
import xml.etree.ElementTree as elemTree
import zipfile
from io import BytesIO

import requests

__author__ = "Yeongbin Jo <iam.yeongbin.jo@gmail.com>"


def get_edgedriver_filename():
    """
    Returns the filename of the binary for the current platform.
    :return: Binary filename
    """
    if sys.platform.startswith("win"):
        return "msedgedriver.exe"
    return "msedgedriver"


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith("win"):
        return ";"
    return ":"


def get_platform_architecture():
    if sys.platform.startswith("linux") and sys.maxsize > 2**32:
        platform = "linux"
        architecture = "64"
    elif sys.platform == "darwin":
        platform = "mac"
        architecture = "64"
    elif sys.platform.startswith("win"):
        platform = "win"
        architecture = "32"
    else:
        raise RuntimeError(
            "Could not determine edgedriver download URL for this platform."
        )
    return platform, architecture


def get_edgedriver_url(version):
    """
    Generates the download URL for current platform , architecture and the given version.
    Supports Linux, MacOS and Windows.
    :param version: edgedriver version string
    :return: Download URL for edgedriver
    """
    base_url = "https://msedgedriver.azureedge.net//"
    platform, architecture = get_platform_architecture()
    return (
        base_url + version + "/edgedriver_" + platform + architecture + ".zip"
    )


def find_binary_in_path(filename):
    """
    Searches for a binary named `filename` in the current PATH. If an executable is found, its absolute path is returned
    else None.
    :param filename: Filename of the binary
    :return: Absolute path or None
    """
    if "PATH" not in os.environ:
        return None
    for directory in os.environ["PATH"].split(get_variable_separator()):
        binary = os.path.abspath(os.path.join(directory, filename))
        if os.path.isfile(binary) and os.access(binary, os.X_OK):
            return binary
    return None


def check_version(binary, required_version):
    try:
        version = subprocess.check_output([binary, "-v"])
        version = re.match(r".*?([\d.]+).*?", version.decode("utf-8"))[1]
        if version == required_version:
            return True
    except Exception:
        return False
    return False


def get_edge_version():
    """
    :return: the version of edge installed on client
    """
    platform, _ = get_platform_architecture()
    if platform == "linux":
        return  # Edge for linux still doesn't exists
    elif platform == "mac":
        process = subprocess.Popen(
            [
                "/Applications/Microsoft Edge.app/Contents/MacOS/Microsoft Edge",
                "--version",
            ],
            stdout=subprocess.PIPE,
        )
        version = (
            process.communicate()[0]
            .decode("UTF-8")
            .replace("Microsoft Edge", "")
            .strip()
        )
    elif platform == "win":
        process = subprocess.Popen(
            [
                "reg",
                "query",
                "HKEY_CURRENT_USER\\Software\\Microsoft\\Edge\\BLBeacon",
                "/v",
                "version",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
        )
        version = process.communicate()[0].decode("UTF-8").strip().split()[-1]
    else:
        return
    return version


def get_major_version(version):
    """
    :param version: the version of edge
    :return: the major version of edge
    """
    return version.split(".")[0]


def get_matched_edgedriver_version(version):
    """
    :param version: the version of edge
    :return: the version of edgedriver
    """
    doc = requests.get(
        "https://msedgedriver.azureedge.net/",
        headers={"accept-encoding": "gzip, deflate, br"},
    ).text
    root = elemTree.fromstring(doc)
    for k in root.iter("Name"):
        if k.text.find(get_major_version(version) + ".") == 0:
            return k.text.split("/")[0]
    return


def get_edgedriver_path():
    """
    :return: path of the edgedriver binary
    """
    return os.path.abspath(os.path.dirname(__file__))


def print_edgedriver_path():
    """
    Print the path of the edgedriver binary.
    """
    print(get_edgedriver_path())


def download_edgedriver(cwd=False):
    """
    Downloads, unzips and installs edgedriver.
    If a edgedriver binary is found in PATH it will be copied, otherwise downloaded.

    :param cwd: Flag indicating whether to download to current working directory
    :return: The file path of edgedriver
    """
    edge_version = get_edge_version()
    if not edge_version:
        logging.debug("edge is not installed.")
        return
    edgedriver_version = get_matched_edgedriver_version(edge_version)
    if not edgedriver_version:
        logging.debug(
            "Can not find edgedriver for currently installed edge version."
        )
        return
    major_version = get_major_version(edgedriver_version)

    if cwd:
        edgedriver_dir = os.path.join(
            os.path.abspath(os.getcwd()), major_version
        )
    else:
        edgedriver_dir = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), major_version
        )
    edgedriver_filename = get_edgedriver_filename()
    edgedriver_filepath = os.path.join(edgedriver_dir, edgedriver_filename)
    if not os.path.isfile(edgedriver_filepath) or not check_version(
        edgedriver_filepath, edgedriver_version
    ):
        logging.debug(f"Downloading edgedriver ({edgedriver_version})...")
        if not os.path.isdir(edgedriver_dir):
            os.makedirs(edgedriver_dir)
        url = get_edgedriver_url(version=edgedriver_version)
        try:
            response = requests.get(url)
            if response.status_code != 200:
                raise Exception("URL Not Found")
        except Exception:
            raise RuntimeError(f"Failed to download edgedriver archive: {url}")
        archive = BytesIO(response.content)
        with zipfile.ZipFile(archive) as zip_file:
            zip_file.extract(edgedriver_filename, edgedriver_dir)
    else:
        logging.debug("edgedriver is already installed.")
    if not os.access(edgedriver_filepath, os.X_OK):
        os.chmod(edgedriver_filepath, 0o744)
    return edgedriver_filepath


if __name__ == "__main__":
    print(get_edge_version())
    print(download_edgedriver())



#
# cucu's browser implementatoin is an abstraction to allow us to easily
# run the same tests against a browser being managed by selenium, cypress or
# any other future browser testing framework
#
import time

from cucu import logger
from cucu.config import CONFIG


class Browser:
    def __init__(self):
        pass

    def open(self, *args, **kwargs):
        raise RuntimeError("implement me")

    def navigate(self, url):
        raise RuntimeError("implement me")

    def switch_to_next_tab(self):
        raise RuntimeError("implement me")

    def switch_to_previous_tab(self):
        raise RuntimeError("implement me")

    def back(self):
        raise RuntimeError("implement me")

    def refresh(self):
        raise RuntimeError("implement me")

    def get_current_url(self):
        raise RuntimeError("implement me")

    def title(self):
        raise RuntimeError("implement me")

    def execute(self, javascript, *args, **kwargs):
        raise RuntimeError("implement me")

    def click(self, element):
        raise RuntimeError("implement me")

    def switch_to_default_frame(self):
        raise RuntimeError("implement me")

    def switch_to_frame(self, frame):
        raise RuntimeError("implement me")

    def screenshot(self, filepath):
        raise RuntimeError("implement me")

    def close_window(self):
        raise RuntimeError("implement me")

    def get_tab_info(self):
        raise RuntimeError("implement me")

    def get_all_tabs_info(self):
        raise RuntimeError("implement me")

    def switch_to_nth_tab(self, tab_index):
        raise RuntimeError("implement me")

    def switch_to_tab_that_matches_regex(self, text):
        raise RuntimeError("implement me")

    def get_session_id(self):
        raise RuntimeError("implement me")

    def quit(self):
        raise RuntimeError("implement me")

    # built in methods to be used by all browser implementations
    def wait_for_page_to_load(self):
        """
        this method is to be used by all browser implementations and called
        when after clicking something or refreshing the page and we'd like to
        make sure the is "loaded" and ready to be interacted with.

        this also gives us a place to fire off any additional page checks such
        as a console log checker, broken image checker, etc.
        """

        # run the page checks
        if CONFIG["__CUCU_PAGE_CHECK_HOOKS"]:
            for name, hook in CONFIG["__CUCU_PAGE_CHECK_HOOKS"].items():
                logger.debug(f'executing page check "{name}"')
                start = time.time()
                hook(self)
                logger.debug(
                    f'executed page check "{name}" in {round(time.time() - start, 3)}s'
                )
